"""This module contains the state information of the heat demand model."""

from datetime import datetime

from pysimmods.util.date_util import GER


class HeatDemandState:
    """Captures the state of the heat demand model"""

    def __init__(self, inits):
        start_date = inits["start_date"]
        if isinstance(start_date, str):
            self.now_dt = datetime.strptime(start_date, GER)
        else:
            self.now_dt = start_date

        self.t_last_3_deg_celsius = inits.get("t_last_3_deg_celsius", 10.3)
        self.t_last_2_deg_celsius = inits.get("t_last_2_deg_celsius", 10.3)
        self.t_last_1_deg_celsius = inits.get("t_last_1_deg_celsius", 10.3)

        self.e_th_kwh = None



"""This module contains the input model for the HVAC model."""

from pysimmods.model.inputs import ModelInputs


class HVACInputs(ModelInputs):
    """Input variables of the HVAC model.

    See :class:`.ModelInputs` for additional information. This class
    has no inputs itself. Instead, each of the values is to be
    provided before each step.

    Attributes
    ----------
    t_air_deg_celsius : float
        Temperature of the environment in [°C].

    """

    def __init__(self):
        super().__init__()

        self.t_air_deg_celsius = None



from __future__ import annotations

import json
from typing import Dict

from .schedule import Schedule


class Flexibilities:
    def __init__(self):
        self._schedules: Dict[str, Schedule] = {}

    def add_schedule(self, schedule_id, schedule):
        self._schedules[f"{schedule_id}"] = schedule

    def __len__(self):
        return len(self._schedules)

    def items(self):
        for key, val in self._schedules.items():
            yield key, val

    def to_json(self):
        d = {key: val.to_dict() for key, val in self._schedules.items()}

        return json.dumps(d)

    def from_json(self, json_str: str) -> Flexibilities:
        self._schedules = {}
        d = json.loads(json_str)

        for key, schedule_dict in d.items():
            self._schedules[key] = Schedule().from_dict(schedule_dict)

        return self

    def __getitem__(self, key):
        return self._schedules[key]



"""This module contains the HVAC model."""

from copy import copy, deepcopy

import numpy as np

from pysimmods.consumer.hvacsim.config import HVACConfig
from pysimmods.consumer.hvacsim.inputs import HVACInputs
from pysimmods.consumer.hvacsim.state import HVACState
from pysimmods.model.consumer import Consumer


class HVAC(Consumer):
    """Simulation model of a heating, ventilation and air conditioning.

    This model is based on a port from the AC model from pratical
    training *energy informatics* of the University of Oldenburg.

    Parameters
    ----------
    params : dict
        Configuration parameters. See :class:`.HVACConfig` for all
        parameters.
    inits : dict
        Initialization parameters. See :class:`.HVACState` for all
        parameters.

    Attributes
    ----------
    config : :class:`.HVACConfig`
        Stores the configuration parameters of the HVAC model.
    state : :class:`.HVACState`
        Stores the initialization parameters of the HVAC model.
    inputs : :class:`.HVACInputs`
        Stores the input parameters for each step of the HVAC model.

    """

    def __init__(self, params, inits):
        self.config = HVACConfig(params)
        self.state = HVACState(inits)
        self.inputs = HVACInputs()

    def step(self):
        """Perform a simulation step."""

        next_state = copy(self.state)
        next_state.p_kw = self.inputs.p_set_kw
        if next_state.p_kw is not None:
            next_state.p_kw = abs(next_state.p_kw)

        self._check_constraints(next_state)

        self._calculate_t(next_state)

        next_state.q_kvar = 0

        self.state = next_state
        self.inputs.reset()

    def _check_constraints(self, next_state):
        if next_state.p_kw is None:
            self._check_internal_temperature(next_state)
        else:
            self._check_schedule(next_state)

    def _check_internal_temperature(self, next_state):
        """Check the temperature constraint

        If internal temperature reaches on the boundaries, cooling is
        activated respectively deactivated, depending on the boundary.

        A cooling HVAC consumes the maximum possible power, a
        non-cooling HVAC consume the minimum possible power (mostly 0).

        """
        if self.state.theta_t_deg_celsius <= self.config.t_min_deg_celsius:
            next_state.cooling = False
        if self.state.theta_t_deg_celsius >= self.config.t_max_deg_celsius:
            next_state.cooling = True

        if next_state.cooling:
            next_state.p_kw = self.config.p_max_kw
        else:
            next_state.p_kw = self.config.p_min_kw

    def _check_schedule(self, next_state):
        """Check if a scheduled operation is possible

        Currently, if p_set_kw was set as input, it will considered
        as one time schedule. If no boundaries are exceeded, the
        model follows the schedule.

        """

        if self.state.theta_t_deg_celsius <= self.config.t_min_deg_celsius:
            next_state.p_kw = self.config.p_min_kw

        elif self.state.theta_t_deg_celsius >= self.config.t_max_deg_celsius:
            next_state.p_kw = self.config.p_max_kw

        else:
            next_state.p_kw = abs(next_state.p_kw)

    def _calculate_t(self, next_state):
        """Calculate the temperature for the next step."""

        minuend = self.config.alpha * (
            self.inputs.t_air_deg_celsius - self.state.theta_t_deg_celsius
        )
        # eta_percent / 100 -> eta decimal
        # p_set_kw * 1000 -> p_set_w
        # -> 1e1
        subtrahend = (
            self.config.eta_percent
            * next_state.p_kw
            * 1e1
            * self.config.cool_factor
        )

        dividend = minuend - subtrahend
        divisor = self.state.mass_kg * self.state.c_j_per_kg_k

        quotient = dividend / divisor
        next_state.theta_t_deg_celsius = self.state.theta_t_deg_celsius + (
            self.inputs.step_size * quotient * self.config.thaw_factor
        )

    def set_percent(self, percentage: float) -> None:
        if percentage is not None and ~np.isnan(percentage):
            return super().set_percent(percentage)

    def get_default_setpoint(self, hour: int) -> float:
        test_state = deepcopy(self.state)
        self._check_constraints(test_state)
        return self._get_percent(
            test_state.p_kw, self.get_pn_min_kw(), self.get_pn_max_kw()
        )



from pysimmods.model.model import Model


class Consumer(Model):
    """A consumer subtype model.

    This class provides unified access to set_percent for all
     generators. A generator returns negative power values in the
     consumer reference arrow system (passive sign convention), which
     is hided if one uses *set_percent*, i.e. calling set_percent
     with, e.g., a value of 50 will always set the power of the
     generator to 50 percent, independently of the reference system
     used.

    """

    def set_p_kw(self, p_kw: float) -> None:
        if p_kw is not None:
            self.inputs.p_set_kw = abs(p_kw)
        else:
            self.inputs.p_set_kw = None

    def get_p_kw(self) -> float:
        return self.state.p_kw * self.config.lsign



"""This module contains the config model for the CHP LPG."""

from pysimmods.model.config import ModelConfig


class CHPLPGConfig(ModelConfig):
    """Config parameters of the CHP LPG.

    This class capturs the configuration parameters of the chp model.

    Parameters
    ----------
    params : dict
        A dictionary containing the configuration parameters.

    Attributes
    ----------
    pn_min_kw : float
        Minimal nominal electrical power output in [kW].
    pn_max_kw : float
        Maximal nominal electrical power output in [kW].
    p_2_p_th_percent : float
        Ratio to compute thermal output from electrical power in [%].
    eta_min_percent : float
        Minimal total efficiency regarding electrical and thermal power
        in [%].
    eta_max_percent : float
        Maximal total efficiency regarding electrical and thermal power
        in [%].
    own_consumption_kw : float
        Own electrical power consumption of the unit [kW].
    active_min_s : int
        Minimal active time of the unit in [s].
    inactive_min_s : int
        Minimal inactive time of the unit in [s].
    lubricant_max_l : float
        Capacity of the lubricant storage in [l].
    lubricant_ml_per_h : float
        Consumption of lubricant per hour in [ml/h].

    """

    def __init__(self, params):
        super().__init__(params)

        self.p_min_kw = abs(params["p_min_kw"])
        self.p_max_kw = abs(params["p_max_kw"])

        self.p_2_p_th_percent = params["p_2_p_th_percent"]
        self.eta_min_percent = params["eta_min_percent"]
        self.eta_max_percent = params["eta_max_percent"]
        self.own_consumption_kw = params["own_consumption_kw"]
        self.active_min_s = params["active_min_s"]
        self.inactive_min_s = params["inactive_min_s"]

        self.lubricant_max_l = params["lubricant_max_l"]
        self.lubricant_ml_per_h = params["lubricant_ml_per_h"]

        self.default_schedule = [
            50,
            50,
            50,
            50,
            50,
            100,
            100,
            100,
            100,
            50,
            50,
            50,
            100,
            100,
            50,
            50,
            50,
            100,
            100,
            100,
            100,
            50,
            50,
            50,
        ]



"""This module contains a Wind turbine."""

import math
from copy import deepcopy

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ...model.generator import Generator
from ...util.air_density_at_height import (
    AIR_DENSITY_KG_PER_M3,
    barometric,
    ideal_gas,
)
from ...util.temperature_at_height import linear_gradient
from ...util.wind_at_height import hellman, logarithmic_profile
from .config import WindPowerPlantConfig
from .inputs import WindPowerPlantInputs
from .state import WindPowerPlantState


class WindPowerPlant(Generator):
    """Simulation model of a windturbine plant.

    The code for the corrections and power output is heavily inspired
    by the windpowerlib::

        https://github.com/wind-python/windpowerlib

    """

    def __init__(self, params, inits):
        self.config: WindPowerPlantConfig = WindPowerPlantConfig(params)
        self.state: WindPowerPlantState = WindPowerPlantState(inits)
        self.inputs: WindPowerPlantInputs = WindPowerPlantInputs()

    def step(self):
        """Perform a simulation step."""
        next_state = deepcopy(self.state)
        self._check_inputs(next_state)

        # First step: wind speed at hub height using hellmann or
        # logarithmic profile
        if self.config.wind_profile == "hellmann":
            next_state.wind_hub_v_m_per_s = hellman(
                self.inputs.wind_v_m_per_s,
                self.config.wind_height_m,
                self.config.hub_height_m,
            )

        elif self.config.wind_profile == "logarithmic":
            next_state.wind_hub_v_m_per_s = logarithmic_profile(
                self.inputs.wind_v_m_per_s,
                self.config.wind_height_m,
                self.config.hub_height_m,
            )

        # Second step: temperature at hub height using linear gradient
        # need T and P  at hub height
        if self.config.temperature_profile == "linear_gradient":
            next_state.t_air_hub_deg_kelvin = linear_gradient(
                self.inputs.t_air_deg_kelvin,
                self.config.temperature_height_m,
                self.config.hub_height_m,
            )
        else:
            next_state.t_air_hub_deg_kelvin = self.inputs.t_air_deg_kelvin

        # Third step: air density at hub height using temperature at
        # hub height. Two different profiles available: barometric or
        # ideal_gas

        if self.config.air_density_profile == "barometric":
            next_state.air_density_hub_kg_per_m3 = barometric(
                self.inputs.air_pressure_hpa,
                next_state.t_air_hub_deg_kelvin,
                self.config.pressure_height_m,
                self.config.hub_height_m,
            )
            wind_speed_corrected = self.wind_curve_correction(
                next_state.air_density_hub_kg_per_m3
            )

        elif self.config.air_density_profile == "ideal_gas":
            next_state.air_density_hub_kg_per_m3 = ideal_gas(
                self.inputs.air_pressure_hpa,
                next_state.t_air_hub_deg_kelvin,
                self.config.pressure_height_m,
                self.config.hub_height_m,
            )
            wind_speed_corrected = self.wind_curve_correction(
                next_state.air_density_hub_kg_per_m3
            )
        else:
            # No air density correction; use default value
            next_state.air_density_hub_kg_per_m3 = AIR_DENSITY_KG_PER_M3
            wind_speed_corrected = self.config.wind_speeds_m_per_s

        # Last step: power output with either power_curve or
        # power_coefficient method
        if self.config.method == "power_curve":
            next_state.p_kw = self.power_curve_output(
                next_state.wind_hub_v_m_per_s, wind_speed_corrected
            )
        elif self.config.method == "power_coefficient":
            next_state.p_kw = self.power_coefficient_output(
                next_state.air_density_hub_kg_per_m3,
                next_state.wind_hub_v_m_per_s,
            )

        self.state = next_state
        self.inputs.reset()

    def wind_curve_correction(
        self, density: float, density0: float = AIR_DENSITY_KG_PER_M3
    ):
        """Power curve at hub height

        Air density initial is set to the default value 1.225
        The correction is displayed in the new values of 
        power_curve_wind_speeds
        """
        interpolation = np.interp(
            self.config.wind_speeds_m_per_s, [7.5, 12.5], [1 / 3, 2 / 3]
        )
        power_curve_correction = (
            np.array(density0 / density).reshape(-1, 1) ** interpolation
        ) * self.config.wind_speeds_m_per_s

        return power_curve_correction[0]

    def power_curve_output(self, wind_speed, power_curve):
        """
        Calculation of the power curve value for a given wind speed by
        interpolation of power curve values dictionary.
        the only value required is the wind speed at hub  height
        wind_hub  = method(logarithmic_profile or hellmann profile)

        """

        power_output = np.interp(
            wind_speed, power_curve, self.config.power_curve_w, left=0, right=0
        )
        return power_output / 1000  # to kW

    def power_coefficient_output(self, air_density, wind_speed):
        """
        calculate the poweroutput using  power coffecient method
        method that requires the following parameters: rotor diameter,  density
        which means it offers more versatile parameters to play with
        """

        power_coefficient_inter = np.interp(
            wind_speed,
            self.config.wind_speeds_m_per_s,
            self.config.power_coefficient,
            left=0,
            right=0,
        )

        return (
            1
            / 8
            * air_density
            * self.config.rotor_diameter**2
            * np.pi
            * np.power(wind_speed, 3)
            * power_coefficient_inter
        ) / 1000  # to kW

    def _check_inputs(self, nstate):
        pass



from pysimmods.model.config import ModelConfig
from pysimmods.model.inputs import ModelInputs
from pysimmods.model.qgenerator import QControl, QGenerator
from pysimmods.model.state import ModelState


class DummyGeneratorConfig(ModelConfig):
    def __init__(self, params):
        super().__init__(params)

        self.p_max_kw = params.get("p_max_kw", 500)
        self.p_min_kw = params.get("p_min_kw", 250)
        self.s_max_kva = params.get("s_max_kva", self.p_max_kw * 1.2)
        self.q_control = params.get("q_control", QControl.PRIORITIZE_P)
        self.cos_phi = params.get("cos_phi", 0.9)


class DummyGeneratorState(ModelState):
    pass


class DummyGeneratorInputs(ModelInputs):
    def __init__(self):
        super().__init__()

        self.cos_phi = None


class DummyQGenerator(QGenerator):
    """The dummy buffer model."""

    def __init__(self, params, inits):
        self.config = DummyGeneratorConfig(params)
        self.state = DummyGeneratorState(inits)
        self.inputs = DummyGeneratorInputs()

    def step(self):
        self.state.p_kw = self.inputs.p_set_kw
        if self.state.p_kw is None:
            self.state.p_kw = 0.0
        if self.state.p_kw != 0:
            self.state.p_kw = min(
                self.config.p_max_kw,
                max(self.config.p_min_kw, self.state.p_kw),
            )

        self.state.q_kvar = self.inputs.q_set_kvar
        if self.state.q_kvar is None:
            self.state.q_kvar = 0.0
        if self.inputs.q_set_kvar != 0:
            self.state.q_kvar = min(
                self.config.s_max_kva,
                max(-self.config.s_max_kva, self.state.q_kvar),
            )



"""This module contains the config model for the chp system."""

from pysimmods.generator.chplpgsim.config import CHPLPGConfig
from pysimmods.model.config import ModelConfig
from pysimmods.other.heatdemandsim.config import HeatDemandConfig


class CHPLPGSystemConfig(ModelConfig):
    """captures the configs for the chp system's components"""

    def __init__(self, params):
        super().__init__(params)

        self.house = HeatDemandConfig(params["household"])
        self.chp = CHPLPGConfig(params["chp"])
        self.e_th_demand_sign = 1.0
        if params.get("flip_e_th_demand_sign", False):
            self.e_th_demand_sign = -1.0
        self.default_schedule = self.chp.default_schedule

    @property
    def p_min_kw(self):
        return self.chp.p_min_kw

    @property
    def p_max_kw(self):
        return self.chp.p_max_kw

    @property
    def q_min_kvar(self):
        return self.chp.q_min_kvar

    @property
    def q_max_kvar(self):
        return self.chp.q_max_kvar



"""This module contains the state model of the chp."""

from pysimmods.model.state import ModelState


class CHPLPGState(ModelState):
    """A CHP LGP state information class

    This class captures the state of the chp model. States normally
    change during the step-method of this model.

    Parameters
    ----------
    inits : dict
        A *dict* containing initialization parameters for the model.

    Attributes
    ----------
    active_s : int
        Time since unit was switched on in [s].
    inactive_s : int
        Time since unit was switch off in [s].
    is_active : bool
        True if the unit is operating.
    storage_t_c : float
        Stores the current temperature of the heat storage in [°C].
    lubricant_l : float, optional
        Current amount of lubricant in [l]. Defaults to 10.

    """

    def __init__(self, inits):
        super().__init__(inits)

        self.active_s = inits["active_s"]
        self.inactive_s = inits["inactive_s"]
        self.is_active = inits["is_active"]
        self.storage_t_c = inits["storage_t_c"]
        self.lubricant_l = inits.get("lubricant_l", 10)



"""This module contains a model of a wind turbine and an inverter"""

from ...model.qgenerator import QGenerator
from ...other.invertersim.inverter import Inverter
from .. import WindPowerPlant
from .config import WindSystemConfig
from .inputs import WindSystemInputs
from .state import WindSystemState


class WindPowerPlantSystem(QGenerator):
    def __init__(self, params, inits):
        self.config = WindSystemConfig(params)
        self.inputs = WindSystemInputs()
        self.state = WindSystemState(inits)

        self.wind = WindPowerPlant(params["wind"], inits["wind"])
        self.inverter = Inverter(params["inverter"], inits["inverter"])

    def step(self):
        """Perform simulation step."""

        # Step the wind turbine
        self.wind.inputs.step_size = self.inputs.step_size
        self.wind.inputs.now_dt = self.inputs.now_dt
        self.wind.inputs.wind_v_m_per_s = self.inputs.wind_v_m_per_s
        self.wind.inputs.t_air_deg_celsius = self.inputs.t_air_deg_celsius
        self.wind.inputs.air_pressure_hpa = self.inputs.air_pressure_hpa
        self.wind.step()

        # Step the inverter
        self.inverter.inputs.p_in_kw = self.wind.state.p_kw
        self.inverter.inputs.p_set_kw = self.inputs.p_set_kw
        self.inverter.inputs.q_set_kvar = self.inputs.q_set_kvar
        self.inverter.inputs.cos_phi_set = self.inputs.cos_phi_set
        self.inverter.inputs.inductive = self.inputs.inverter_inductive
        self.inverter.step()

        # Update state
        self.state.p_kw = self.inverter.state.p_kw
        self.state.p_possible_max_kw = self.wind.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi
        self.state.inverter_inductive = self.inverter.state.inductive
        self.state.wind_hub_v_m_per_s = self.wind.state.wind_hub_v_m_per_s
        self.state.t_air_hub_deg_kelvin = self.wind.state.t_air_hub_deg_kelvin
        self.state.air_density_hub_kg_per_m3 = (
            self.wind.state.air_density_hub_kg_per_m3
        )

        self.inputs.reset()

    def get_state(self):
        state_dict = {
            "wind": self.wind.get_state(),
            "inverter": self.inverter.get_state(),
        }
        return state_dict

    def set_state(self, state_dict):
        self.wind.set_state(state_dict["wind"])
        self.inverter.set_state(state_dict["inverter"])

        self.state.p_kw = self.inverter.state.p_kw
        self.state.p_possible_max_kw = self.wind.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi
        self.state.inverter_inductive = self.inverter.state.inductive
        self.state.wind_hub_v_m_per_s = self.wind.state.wind_hub_v_m_per_s
        self.state.t_air_hub_deg_kelvin = self.wind.state.t_air_hub_deg_kelvin
        self.state.air_density_hub_kg_per_m3 = (
            self.wind.state.air_density_hub_kg_per_m3
        )

    def set_q_kvar(self, q_kvar: float):
        self.inputs.q_set_kvar = q_kvar



import numpy as np


def logarithmic_profile(
    wind_speed,
    wind_speed_height,
    hub_height,
    roughness_length=0.15,
    obstacle_height=0.0,
):
    return (
        wind_speed
        * np.log((hub_height - 0.7 * obstacle_height) / roughness_length)
        / np.log(
            (wind_speed_height - 0.7 * obstacle_height) / roughness_length
        )
    )


# Calculate the wind speed at altitude h using hellmann formula
def hellman(wind_speed, wind_speed_height, hub_height, hellman_exponent=1 / 7):
    return wind_speed * (hub_height / wind_speed_height) ** hellman_exponent



"""This module contains the config model for the Wind plant."""

import numpy as np

from ...model.config import ModelConfig


class WindPowerPlantConfig(ModelConfig):
    """Config parameters of Wind plant model.

    Parameters
    ----------

    params : dict
        Contains the configuration of the turbine. See attribute
        section for more information about the parameters, attributes
        marked with *(Input)* can or must be provided.

    Attributes
    ----------

    wind_speed_height:
    roughness_length:
    obstacle_height:
    air_density:
        represent the default value
    nominal_power:
    hub_height:
    rotor_diameter:
    turbine_type:
    pressure:
    temperature:
    temperature_height:
    pressure:

    """

    def __init__(self, params):
        super().__init__(params)

        self.p_max_kw = params.get("pn_max_kw", 2000)
        self.p_min_kw = 0
        self.turbine_type = params.get("turbine_type", "E-82/2000")
        self.hub_height_m = params.get("hub_height_m", 78)
        self.rotor_diameter_m = params.get("rotor_diameter_m", 82)
        self.roughness_length_m = params.get("roughness_length_m", 15)
        self.obstacle_height_m = params.get("obstacle_height_m", 0)

        self.wind_height_m = params.get("wind_height_m", 10)
        self.temperature_height_m = params.get("temperature_height_m", 10)
        self.pressure_height_m = params.get("pressure_height_m", 10)
        # self.air_density = params.get("air_density", 1.225)
        # self.nominal_power = params.get("nominal_power", 10e6)

        # self.pressure = params.get("pressure", 98405.7)
        # self.temperature = params.get("temperature", 268)

        self.method = params.get("method", "power_curve")

        self.temperature_profile = params.get(
            "temperature_profile", "linear_gradient"
        )
        if self.temperature_profile != "linear_gradient":
            self.temperature_profile = "no_profile"

        # self.pressure_profile = params.get(
        #     "pressure_profile", "atmospheric_pressure"
        # )

        self.wind_profile = params.get("wind_profile", "hellmann")

        self.air_density_profile = params.get(
            "air_density_profile", "barometric"
        )
        self.power_curve_w = np.array(
            params.get("power_curve_m", [0, 26, 180, 1500, 3000, 3000])
        )

        self.power_coefficient = np.array(
            params.get(
                "power_coefficient", [0.301, 0.36, 0.39, 0.409, 0.421, 0.429]
            )
        )

        self.wind_speeds_m_per_s = np.array(params.get("wind_speeds", []))
        if self.wind_speeds_m_per_s.size == 0:
            if self.method == "power_curve":
                num_vals = len(self.power_curve_w)
            else:
                num_vals = len(self.power_coefficient)

            self.wind_speeds_m_per_s = np.linspace(0.0, 25.0, num_vals)

        # power_curve_wind_speed = params.get(
        #     "power_curve_wind_speeds", [0.0, 2.0, 4.0, 6.0, 8.0, 10.0]
        # )

        # self.power_curve = pd.DataFrame(
        #     data={
        #         "value": power_curve_values,
        #         "wind_speed": power_curve_wind_speed,
        #     }
        # )
        # power_coefficient_wind_speed = params.get(
        #     "power_coefficient_curve_wind_speeds",
        #     [4.0, 4.5, 5.0, 5.5, 6.0, 6.5],
        # )
        # power_coefficient_curve_value = params.get(
        #     "power_coefficient_curve_values",
        #     [0.301, 0.36, 0.39, 0.409, 0.421, 0.429],
        # )
        # power_coefficient_dict = {
        #     "value": power_coefficient_curve_value,
        #     "wind_speed": power_coefficient_wind_speed,
        # }
        # self.power_coefficient = pd.DataFrame(power_coefficient_dict)

        # self.p_min_kw = 0
        # self.p_max_kw = power_curve_wind_speed[-1]
        # self.default_p_schedule = None
        # self.default_q_schedule = None



"""This module contains the input model for the chp."""

from pysimmods.model.inputs import ModelInputs


class CHPLPGInputs(ModelInputs):
    """A CHP input class

    This class captures the inputs for the chp model. The values need to be
    provided in before calling the step method. However, once an input is set
    it will be reused as long as it is not changed.

    Attributes
    ----------
    e_th_demand_set_kwh : float
        The setpoint for thermal energy demand in [kWh].
    """

    def __init__(self):
        super().__init__()

        self.e_th_demand_set_kwh = None



from pysimmods.model.buffer import Buffer
from pysimmods.model.config import ModelConfig
from pysimmods.model.inputs import ModelInputs
from pysimmods.model.state import ModelState


class DummyBufferConfig(ModelConfig):
    def __init__(self, params):
        params.setdefault("default_schedule", [10] * 24)
        super().__init__(params)

        self.p_max_kw = params.get("p_max_kw", 500)
        self.p_min_kw = params.get("p_min_kw", 250)
        self.q_max_kvar = params.get("q_max_kvar", 100)
        self.q_min_kvar = params.get("q_min_kvar", 25)

        if "p_charge_max_kw" in params:
            self.p_charge_max_kw = params["p_charge_max_kw"]
        if "p_charge_min_kw" in params:
            self.p_charge_min_kw = params["p_charge_min_kw"]
        if "p_discharge_max_kw" in params:
            self.p_discharge_max_kw = params["p_discharge_max_kw"]
        if "p_discharge_min_kw" in params:
            self.p_discharge_min_kw = params["p_discharge_min_kw"]


class DummyBufferState(ModelState):
    pass


class DummyBufferInputs(ModelInputs):
    pass


class DummyBuffer(Buffer):
    """The dummy buffer model."""

    def __init__(self, params, inits):
        self.config = DummyBufferConfig(params)
        self.state = DummyBufferState(inits)
        self.inputs = DummyBufferInputs()

    def step(self):
        self.state.p_kw = self.inputs.p_set_kw



"""This module contains the python port of a CHP model developed in the
student's project group POWDER.
"""

from copy import copy

from pysimmods.generator.chplpgsim.config import CHPLPGConfig
from pysimmods.generator.chplpgsim.inputs import CHPLPGInputs
from pysimmods.generator.chplpgsim.state import CHPLPGState
from pysimmods.model.generator import Generator
from pysimmods.other.heatstoragesim.heatstorage import HeatStorage


class CHPLPG(Generator):
    """A Combined Heat and Power unit model.

    The CHP is fueled with Liquefied Petroleum Gas (LPG).

    Attributes
    ----------
    config : CHPLPGConfig
    state : CHPLPGState
    inputs : CHPLPGInputs

    """

    def __init__(self, params, inits):
        self.config = CHPLPGConfig(params)
        self.state = CHPLPGState(inits)
        self.inputs: CHPLPGInputs = CHPLPGInputs()

        self._storage = HeatStorage(params, inits)

    def step(self):
        """Perform a simulation step."""
        next_state = copy(self.state)
        next_state.p_kw = self._get_setpoint()

        self._check_inputs()

        # Operating state constraint is checked first to prevent
        # invalid user inputs. However, this has least priority and
        # changes due to other constraints are possible.
        self._check_operating_state(next_state)

        self._check_performance_limit(next_state)
        self._check_lubricant(next_state)
        self._check_storage_temperature(next_state)

        # Change operating state
        if next_state.p_kw != 0:
            next_state.is_active = True
            next_state.active_s += self.inputs.step_size
            next_state.inactive_s = 0
        else:
            next_state.is_active = False
            next_state.active_s = 0
            next_state.inactive_s += self.inputs.step_size
            # Auto refill of lubricant
            next_state.lubricant_l = self.config.lubricant_max_l

        next_state.q_kvar = 0

        self.state = next_state
        self.inputs.reset()

    def _check_inputs(self):
        if self.inputs.e_th_demand_set_kwh is None:
            self.inputs.e_th_demand_set_kwh = 0

    def _get_setpoint(self):
        setpoint = self.inputs.p_set_kw

        if setpoint is not None:
            return abs(setpoint)

        hour = self.inputs.now_dt.hour
        default = self.config.default_schedule[hour]
        setpoint = self.config.p_max_kw * default / 100

        return abs(setpoint)

    def _check_operating_state(self, next_state):
        """Check if operating state constraints are satisfied.

        If the inut is not allowed to switch on, it will stay off.
        On the other hand, if the unit is not allowed to switch off,
        the minimal power is used as set value.

        """

        if next_state.p_kw > 0:
            if (
                not self.state.is_active
                and self.state.inactive_s < self.config.inactive_min_s
            ):
                next_state.p_kw = 0
        else:
            if (
                self.state.is_active
                and self.state.active_s < self.config.active_min_s
            ):
                next_state.p_kw = self.config.p_min_kw

    def _check_performance_limit(self, next_state):
        """Check if minimal performance is reached.

        If this is not the case, the set value is adapted to minimal
        power.

        """
        if next_state.p_kw > 0:
            next_state.p_kw = max(
                min(next_state.p_kw, self.config.p_max_kw),
                self.config.p_min_kw,
            )

    def _check_lubricant(self, next_state):
        """Check if lubricant constraint is satisfied.

        If no lubricant is available, the unit will switch off.

        """
        if next_state.p_kw > 0:
            lubricant_delta_l = (
                self.config.lubricant_ml_per_h
                * (self.inputs.step_size / 3_600)
                / 1_000
            )
            if self.state.lubricant_l < lubricant_delta_l:
                # not enough lubricant, switch off
                next_state.p_kw = 0
            else:
                next_state.lubricant_l -= lubricant_delta_l

    def _check_storage_temperature(self, next_state):
        """Check if the storage temperature is not too high

        Also calculates the new storage temperature and if the
        temperature is too high, a lower set value for power is
        calculated.

        """
        # Thermal energy production
        e_th_prod_kwh = (
            next_state.p_kw
            * self.config.p_2_p_th_percent
            / 100
            * self.inputs.step_size
            / 3_600
        )

        # Get maximal energy the storage can absorb
        e_th_in_max_kwh = self._storage.get_absorbable_energy()

        if e_th_prod_kwh > e_th_in_max_kwh:
            next_state.p_kw = (
                e_th_in_max_kwh
                * (100 / self.config.p_2_p_th_percent)
                * (3_600 / self.inputs.step_size)
            )
            e_th_prod_kwh = e_th_in_max_kwh

        # Get available storage energy
        e_th_in_min_kwh = self._storage.get_available_energy()
        demand_after_prod = min(
            0, self.inputs.e_th_demand_set_kwh + e_th_prod_kwh
        )

        if abs(demand_after_prod) > e_th_in_min_kwh:
            next_state.p_kw = max(
                self.config.p_min_kw,
                next_state.p_kw + (self.config.p_min_kw * 0.25),
            )
            return self._check_storage_temperature(next_state)

        self._storage.absorb_energy(
            e_th_prod_kwh, self.inputs.e_th_demand_set_kwh
        )

        next_state.storage_t_c = self._storage.t_c
        next_state.p_th_kw = (
            next_state.p_kw * self.config.p_2_p_th_percent / 100
        )

    def set_state(self, state_dict):
        """Set the state."""
        super().set_state(state_dict)
        self._storage.t_c = self.state.storage_t_c

    # @property
    # def set_percent(self):
    #     p_kw = self.inputs.p_set_kw
    #     if p_kw is None:
    #         p_kw = self.state.p_kw

    #     p_kw = (p_kw - self.config.pn_min_kw) / (
    #         self.config.pn_max_kw - self.config.pn_min_kw
    #     )
    #     return abs(p_kw) * 100

    # @set_percent.setter
    # def set_percent(self, value):
    #     value = max(min(abs(value), 100.0), 0.0)

    #     if value == 0:
    #         p_set_kw = 0
    #     else:
    #         p_range = self.config.pn_max_kw - self.config.pn_min_kw
    #         p_set_kw = self.config.pn_min_kw + value / 100 * p_range

    #     self.inputs.p_set_kw = p_set_kw



"""This module contains the state model of the Wind Turbine System."""

from ...model.state import ModelState


class WindSystemState(ModelState):
    def __init__(self, inits):
        super().__init__(inits)

        self.p_possible_max_kw: float = 0
        self.wind_hub_v_m_per_s: float = 0
        self.t_air_hub_deg_kelvin: float = 0
        self.air_density_hub_kg_per_m3: float = 0



from typing import Optional

from .model import Model


class Generator(Model):
    """A generator subtype model.

    A generator returns negative power values in the
    consumer reference arrow system.

    """

    def set_p_kw(self, p_kw: Optional[float]) -> None:
        if p_kw is not None:
            self.inputs.p_set_kw = abs(p_kw)
        else:
            self.inputs.p_set_kw = None

    def get_p_kw(self) -> float:
        return self.state.p_kw * self.config.gsign

    def get_qn_min_kvar(self) -> float:
        return 0

    def get_qn_max_kvar(self) -> float:
        return 0



from datetime import datetime
from typing import Optional, Tuple

import numpy as np

from pysimmods.model.model import Model
from pysimmods.other.flexibility.schedule import Schedule


class ScheduleModel:
    """A wrapper for pysimmods, which allows models to use schedules.

    Parameters
    ----------
    model: :class:`.Model`
        The model that should be wrapped by the schedule model.
    unit: str, optional
        The unit to be used by the schedule. Default is `kw` for kilo
        watt. Other options are `w` and `mw`.
    prioritize_setpoint: bool, optional
        If set to True, this model will prefer setpoints rather than
        schedule values if setpoints are present. Default is False.
    """

    def __init__(
        self,
        model: Model,
        unit: str = "kw",
        prioritize_setpoint: bool = False,
        step_size: Optional[int] = None,
        now_dt: Optional[datetime] = None,
    ):
        self._model = model
        self._prioritize_setpoint = prioritize_setpoint

        if unit == "mw":
            self._unit_factor = 1e-3
            unit_str = "m"

        elif unit == "w":
            self._unit_factor = 1e3
            unit_str = ""
        else:
            self._unit_factor = 1
            unit_str = "k"

        self._pname = f"p_{unit_str}w"
        self._qname = f"q_{unit_str}var"
        self._psetname = f"p_set_{unit_str}w"
        self._qsetname = f"q_set_{unit_str}var"

        self.schedule: Schedule = Schedule()
        self._step_size: Optional[float] = step_size
        self._now_dt: Optional[datetime] = now_dt
        self._percent_factor: float

        if self._model.config.use_decimal_percent:
            self._percent_factor = 0.01
        else:
            self._percent_factor = 1.0

    def update_schedule(self, schedule):
        if self.schedule.is_empty():
            self._check_inputs()
        self.schedule.update(schedule)

    def step(self):
        """Perform a simulation step of the underlying model."""

        self._check_inputs()

        p_set, q_set = self._get_setpoints()
        self._model.set_p_kw(p_set)
        self._model.set_q_kvar(q_set)

        self._model.step()

        if p_set is None:
            p_set = self._model.get_p_kw()
        if q_set is None:
            q_set = self._model.get_q_kvar()

        self.schedule.update_row(
            self._now_dt,
            p_set * self._unit_factor,
            q_set * self._unit_factor,
            self._model.get_p_kw() * self._unit_factor,
            self._model.get_q_kvar() * self._unit_factor,
        )
        self.schedule.now_dt = self._now_dt
        self.schedule.prune()

    def _check_inputs(self) -> None:
        # Check if model has a different step size
        if (
            self._model.inputs.step_size is not None
            and self._model.inputs.step_size != self._step_size
        ):
            self._step_size = self._model.inputs.step_size

        # Check if model uses a different time
        if (
            self._model.inputs.now_dt is not None
            and self._model.inputs.now_dt != self._now_dt
        ):
            self._now_dt = self._model.inputs.now_dt

        # Initialize schedule if necessary
        if self.schedule.is_empty():
            self.schedule.p_name = self._pname
            self.schedule.q_name = self._qname
            self.schedule.now_dt = self._now_dt
            self.schedule.step_size = self._step_size
            self.schedule.p_set_name = self._psetname
            self.schedule.q_set_name = self._qsetname

            self.schedule.horizon_hours = self._step_size / 3_600
            self.schedule.init()

    def _get_setpoints(self) -> Tuple[Optional[float], Optional[float]]:
        try:
            schedule_p = (
                self.schedule.get(self._now_dt, self._psetname)
                / self._unit_factor
            )
        except TypeError:
            schedule_p = None
        try:
            schedule_q = (
                self.schedule.get(self._now_dt, self._qsetname)
                / self._unit_factor
            )
        except TypeError:
            schedule_q = None

        model_p = self._model.get_p_set_kw()
        model_q = self._model.get_q_set_kvar()

        default_p = self._model.get_default_p_set(self._now_dt.hour)
        default_q = self._model.get_default_q_set(self._now_dt.hour)

        priority = [(schedule_p, schedule_q), (model_p, model_q)]
        if self._prioritize_setpoint:
            priority = priority[::-1]
        priority.append((default_p, default_q))

        setpoint = (None, None)
        for setvals in priority[::-1]:
            if setvals[0] is not None and ~np.isnan(setvals[0]):
                setpoint = (setvals[0], setpoint[1])
            if setvals[1] is not None and ~np.isnan(setvals[1]):
                setpoint = (setpoint[0], setvals[1])

        return setpoint

    def set_step_size(self, step_size):
        self._model.set_step_size(step_size)
        self._step_size = self._model.inputs.step_size

    def set_now_dt(self, now_dt):
        self._model.set_now_dt(now_dt)
        self._now_dt = self._model.inputs.now_dt

    def set_p_kw(self, p_kw):
        self._model.set_p_kw(p_kw)

    def set_q_kvar(self, q_kvar):
        self._model.set_q_kvar(q_kvar)

    def get_now_dt(self):
        return self._model.get_now_dt()

    def get_p_kw(self):
        return self._model.get_p_kw()

    def get_q_kvar(self):
        return self._model.get_q_kvar()

    def get_pn_max_kw(self):
        return self._model.get_pn_max_kw()

    def get_pn_min_kw(self):
        return self._model.get_pn_min_kw()

    def get_qn_max_kvar(self):
        return self._model.get_qn_max_kvar()

    def get_qn_min_kvar(self):
        return self._model.get_qn_min_kvar()

    @property
    def inputs(self):
        return self._model.inputs

    @property
    def config(self):
        return self._model.config

    @property
    def state(self):
        return self._model.state



"""This module contains the state model of the inverter."""

from pysimmods.model.state import ModelState


class InverterState(ModelState):
    """Inverter state

    Parameters
    ----------
    inits : dict
        A *dict* containing initialization parameters of the inverter.

    Attributes
    ----------
    cos_phi : float
        The cosinus of the phase angle used in the last step.
    inductive: float
        Indicates whether the inverter used inductive (True) or
        capacitive (False) mode in the last step.
    """

    def __init__(self, inits):
        super().__init__(inits)

        self.cos_phi = inits.get("cos_phi", 0.9)
        self._inductive = False

    @property
    def inductive(self):
        if self._inductive:
            return 1
        else:
            return 0



"""This module contains the config of the Wind Turbine System."""

from ...model.config import ModelConfig
from ...other.invertersim.config import InverterConfig
from ..windsim.config import WindPowerPlantConfig


class WindSystemConfig(ModelConfig):
    def __init__(self, params):
        super().__init__(params)
        params["wind"]["sign_convention"] = self.sign_convention
        params["inverter"]["sign_convention"] = self.sign_convention

        # Those are duplicate to the original configs
        self._turbine: WindPowerPlantConfig = WindPowerPlantConfig(
            params["wind"]
        )
        self._inverter: InverterConfig = InverterConfig(params["inverter"])

        self.default_p_schedule = None
        self.default_q_schedule = None

    @property
    def s_max_kva(self):
        return self._inverter.s_max_kva

    @property
    def q_control(self):
        return self._inverter.q_control

    @property
    def cos_phi(self):
        return self._inverter.cos_phi

    @property
    def p_max_kw(self):
        return self._turbine.p_max_kw

    @property
    def p_min_kw(self):
        return self._turbine.p_min_kw



from ..windsim.presets import wind_power_plant_preset


def wind_presets(
    pn_max_kw,
    wind_profile: str = "hellmann",
    temperature_profile: str = "linear_gradient",
    air_density_profile: str = "barometric",
    power_method: str = "power_curve",
    rng=None,
    turbine_type=None,
    force_type=False,
    cos_phi=0.9,
    q_control="prioritize_p",
    inverter_mode="inductive",
    **kwargs,
):
    # print(pn_max_kw, cos_phi, q_control, inverter_mode, kwargs)
    wparams, winit = wind_power_plant_preset(
        pn_max_kw,
        wind_profile,
        temperature_profile,
        air_density_profile,
        power_method,
        rng,
        turbine_type,
        force_type,
        **kwargs,
    )

    params = {
        "wind": wparams,
        "inverter": {
            "sn_kva": pn_max_kw / cos_phi,
            "q_control": q_control,
            "cos_phi": cos_phi,
            "inverter_mode": inverter_mode,
        },
        "sign_convention": "active",
    }

    inits = {"wind": winit, "inverter": None}

    return params, inits



"""This module contains the config model of the PV plant system."""

from pysimmods.generator.pvsim.config import PVConfig
from pysimmods.model.config import ModelConfig
from pysimmods.other.invertersim.config import InverterConfig


class PVSystemConfig(ModelConfig):
    """Config parameters of the PV plant System.

    Consists of a :class:`~.PVConfig` and an :class:`~.InverterConfig`
    object.

    """

    def __init__(self, params):
        super().__init__(params)
        params["pv"]["sign_convention"] = self.sign_convention
        params["inverter"]["sign_convention"] = self.sign_convention

        self.pv = PVConfig(params["pv"])
        self.inverter = InverterConfig(params["inverter"])
        self.default_p_schedule = None
        self.default_q_schedule = None

    @property
    def s_max_kva(self):
        return self.inverter.s_max_kva

    @property
    def q_control(self):
        return self.inverter.q_control

    @property
    def cos_phi(self):
        return self.inverter.cos_phi

    @property
    def p_max_kw(self):
        return self.pv.p_max_kw

    @property
    def p_min_kw(self):
        return self.pv.p_min_kw

    # @property
    # def p_min_kw(self):
    #     """Minimum active power of pv system in kW, this property is
    #     used by :class:`.mosaik.flexibility_model.FlexibilityModel`.

    #     With passive sign convention, this is the maximum power this
    #     plant can generate.

    #     """
    #     if self.psc:
    #         return self._get_abs_p_max()
    #     else:
    #         return 0

    # @property
    # def p_max_kw(self):
    #     """Maximum active power of pv plant in kW, this property is
    #     used by :class:`.mosaik.flexibility_model.FlexibilityModel`.

    #     With active sign convention, this is the maximum power this
    #     plant can generate

    #     """
    #     if self.psc:
    #         return 0
    #     else:
    #         return self._get_abs_p_max()

    # @property
    # def q_min_kvar(self):
    #     """Minimum reactive power of pv plant in kVAr.

    #     For certain combinations, this is the maximum absolute power
    #     this plant can generate.

    #     """
    #     if self.psc and self.inverter.inverter_mode == "capacitive":
    #         return self._get_abs_q_max()
    #     elif self.asc and self.inverter.inverter_mode == "inductive":
    #         return self._get_abs_q_max()
    #     else:
    #         return 0

    # @property
    # def q_max_kvar(self):
    #     """Maximum reactive power of pv plant in kVAr.

    #     For certain combinations, this is the maximum absolute power
    #     this plant can generate.

    #     """
    #     if self.psc and self.inverter.inverter_mode == "inductive":
    #         return self._get_abs_q_max()
    #     elif self.asc and self.inverter.inverter_mode == "capacitive":
    #         return self._get_abs_q_max()
    #     else:
    #         return 0

    # @property
    # def default_schedule(self):
    #     p_dominated = ["p_set", "pq_set"]
    #     q_dominated = ["q_set", "qp_set"]
    #     cos_phi_dominated = ["cos_phi_set"]
    #     mode = self.inverter.q_control

    #     if mode in p_dominated:
    #         return self.pv.default_schedule
    #     elif mode in q_dominated:
    #         return [20 for _ in range(24)]
    #     elif mode in cos_phi_dominated:
    #         return [90 for _ in range(24)]
    #     else:
    #         return self.pv.default_schedule

    # def _get_abs_p_max(self):
    #     if self.pv.p_peak_kw > self.inverter.sn_kva:
    #         return self.inverter.sn_kva * self.gsign
    #     else:
    #         return self.pv.p_peak_kw * self.gsign

    # def _get_abs_q_max(self):
    #     const_cos_phi_modes = ["p_set", "q_set"]
    #     variable_cos_phi_modes = ["cos_phi_set", "pq_set", "qp_set"]

    #     if self.inverter.q_control in const_cos_phi_modes:
    #         q_max_kvar = self.inverter.sn_kva * math.sin(
    #             math.acos(self.inverter.cos_phi)
    #         )

    #         return q_max_kvar * self.gsign

    #     elif self.inverter.q_control in variable_cos_phi_modes:
    #         return self.inverter.sn_kva * self.gsign

    #     else:
    #         return 0



"""This module contains the contains the config model for HVAC."""

from pysimmods.model.config import ModelConfig


class HVACConfig(ModelConfig):
    """Config parameters of the HVAC model.

    Parameters
    ----------
    params : dict
        Contains the configuration for the HVAC model. See attribute
        section for more information about the parameters, attributes
        marked with *(Input)* can or must be provided.

    Attributes
    ----------
    pn_min_kw : float
        (Input) Nominal minimal power output in [kW].
    pn_max_kw : float
        (Input) Nominal power output in [kW].
    eta_percent : float
        (Input) Efficiency of the model in [%].
    l_m : float
        (Input) The length of the room to be cooled in [m].
    w_m : float
        (Input) The width of the room to be cooled in [m].
    h_m : float
        (Input) The height of the room to be cooled in [m].
    v_m3 : float
        The volume of the room to be cooled in [m³]. Is calculated
        from l_m, w_m, and h_m.
    a_m2 : float
        Surface of the room to be cooled in [m²]. Is calculated from
        l_m, w_m, and h_m.
    d_m : float
        Thickness of isolation in [m].
    lambda_w_per_m_k : float
        (Input) Thermal conductivity of isolation in [W*m^-1*K^-1].
    alpha : float
        Calculated from lambda_w_per_m_k, a_m2, and d_m.
    t_min_deg_celsius : float
        (Input) When this temperature is reached, the HVAC starts
        cooling, in [°C].
    t_max_deg_celsius : float
        (Input) When this temperature is reached, the HVAC stops
        cooling, in [°C].
    thaw_factor : float, optional
        Extra factor to control the speed of the thawing process.
        Defaults to 1.0.
    cool_factor : float, optional
        Extra factor control the speed of the cooling process.
        Defaults to 1.0.

    """

    def __init__(self, params):
        super().__init__(params)

        self.p_min_kw = 0
        self.p_max_kw = params.get("p_max_kw", 2)
        self.eta_percent = params.get("eta_percent", 200.0)
        self.l_m = params.get("l_m", 4)
        self.w_m = params.get("w_m", 5)
        self.h_m = params.get("h_m", 2.5)
        self.d_m = params.get("d_m", 0.25)
        self.lambda_w_per_m_k = params.get("lambda_w_per_m_k", 0.5)
        self.t_min_deg_celsius = params.get("t_min_deg_celsius", 17)
        self.t_max_deg_celsius = params.get("t_max_deg_celsius", 23)

        self.a_m2 = (
            self.l_m * self.w_m + self.l_m * self.h_m + self.w_m * self.h_m
        ) * 2
        self.v_m3 = self.l_m * self.w_m * self.h_m
        self.alpha = self.lambda_w_per_m_k * self.a_m2 / self.d_m

        self.thaw_factor = params.get("thaw_factor", 1.0)
        self.cool_factor = params.get("cool_factor", 1.0)

        # Model follows cooling demand;
        # Use ScheduleModel to set schedules
        self.default_schedule = None

        # self.default_schedule = list()
        # for _ in range(8):
        #     self.default_schedule.extend([80, 10, 10])



"""This module contains an inverter model"""

from copy import copy
from math import sqrt

from pysimmods.model.qgenerator import QControl, QGenerator
from pysimmods.other.invertersim.config import InverterConfig
from pysimmods.other.invertersim.inputs import InverterInputs
from pysimmods.other.invertersim.state import InverterState


class Inverter(QGenerator):
    """Inverter model

    Supports five different modes:

    - *p_set*: p_set_kw is provided, cos phi remains constant and
      q_kvar is calculated with sqrt(s^2-p^2).
    - *q_set*: q_set_kvar is provided, cos phi remains constant and
      p_kw is calculated with sqrt(s^2-q^2)
    - *cos_phi_set*: cos phi is provided, p_kw is calculated from
      p_in_kw and q_kvar is calculated accordingly.
    - *pq_set*: p_set_kw and q_set_kvar are provided, with p being
      prioritized, i.e., q is limited by sqrt(s^2-p^2), and cos phi is
    calculated with p/s.
    - *qp_set*: q_set_kvar and p_set_kw are provided, with q being
      prioritized, i.e., p is limited by sqrt(s^2-q^2), and cos phi is
      calculated with p/s

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters.
    inits : dict, optional
        A *dict* containing initialization parameters.

    Attributes
    ----------
    config : :class:`~.InverterConfig`
        Configuration parameters of the inverter.
    state : :class:`~.InverterState`
        Initialization parameters of the inverter.
    inputs : :class:`~.InverterInputs`
        Input parameters of the inverter.

    """

    def __init__(self, params, inits=None):
        if inits is None:
            inits = {}

        self.config: InverterConfig = InverterConfig(params)
        self.state: InverterState = InverterState(inits)
        self.inputs: InverterInputs = InverterInputs()

    def step(self):
        """Perform simulation step"""
        self._check_inputs()

        next_state = copy(self.state)
        next_state._inductive = self.inputs._inductive
        self._calculate(next_state)

        self.state = next_state
        self.inputs.reset()

    def _check_inputs(self):
        if self.inputs.p_set_kw is None:
            self.inputs.p_set_kw = self.inputs.p_in_kw

        if self.inputs.cos_phi_set is None:
            self.inputs.cos_phi_set = self.config.cos_phi

        if self.inputs._inductive is None:
            self.inputs._inductive = self.config.inverter_mode == "inductive"

    def _calculate(self, next_state: InverterState):
        # Can't output more power than is available
        p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

        # Check constraints for apparent power
        s_kva = p_kw / self.inputs.cos_phi_set
        s_kva = min(s_kva, self.config.s_max_kva)

        # Use updated apparent power to calculate active power
        p_max_kw = s_kva * self.inputs.cos_phi_set
        p_kw = min(p_kw, p_max_kw)

        # Calculate maximum reactive power
        q_max_kvar = sqrt(self.config.s_max_kva**2 - p_kw**2)

        # Check if reactive power setpoint is provided
        if self.inputs.q_set_kvar is None:
            # No; use "remaining" current apparent power for reactive
            # power, preserving the cos phi
            q_kvar = sqrt(s_kva**2 - p_kw**2)
            q_kvar = min(q_max_kvar, q_kvar)
            if next_state._inductive:
                q_kvar *= -1
        else:
            # Yes; use "remaining" maximum apparent power for reactive
            # power, the cos phi may change
            q_set_sign = 1 if self.inputs.q_set_kvar >= 0 else -1
            q_kvar = min(
                abs(self.inputs.q_set_kvar), abs(self.config.s_max_kva)
            )
            s_kva = sqrt(p_kw**2 + q_kvar**2)

            # Limit active or reactive power according to the q-control
            # mode
            if s_kva > self.config.s_max_kva:
                if self.config.q_control == QControl.PRIORITIZE_P:
                    q_kvar = q_max_kvar
                elif self.config.q_control == QControl.PRIORITIZE_Q:
                    p_kw = sqrt(self.config.s_max_kva**2 - q_kvar**2)

            q_kvar *= q_set_sign

        if s_kva > 0:
            next_state.cos_phi = p_kw / s_kva
        else:
            next_state.cos_phi = self.inputs.cos_phi_set
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar

    def set_q_kvar(self, q_kvar: float) -> None:
        self.inputs.q_set_kvar = q_kvar



"""
This module contains a model of a chp system with a chp and a
household. The household has a heat demand which allows the chp
to run for at least 5000 hours in a year.

"""

import logging

from pysimmods.generator.chplpgsim import CHPLPG
from pysimmods.model.generator import Generator
from pysimmods.other.heatdemandsim.heatdemand import create_heatdemand

from .config import CHPLPGSystemConfig
from .inputs import CHPLPGSystemInputs
from .state import CHPLPGSystemState

LOG = logging.getLogger(__name__)


class CHPLPGSystem(Generator):
    """CHP system with CHP and household"""

    def __init__(self, params, inits):
        # First, create the chp
        params["chp"]["sign_convention"] = params.get(
            "sign_convention", "passive"
        )
        self.chp = CHPLPG(params["chp"], inits["chp"])

        # Get the thermal capabilities of the chp
        p_th_min_kw = (
            self.chp.config.p_min_kw * self.chp.config.p_2_p_th_percent * 0.01
        )

        params.setdefault("household", dict())
        params["household"]["chp_p_th_prod_kw"] = p_th_min_kw

        # Second, create an appropriate household
        self.heatdemand = create_heatdemand(
            params["household"], inits["household"]
        )

        self.config = CHPLPGSystemConfig(params)
        self.state = CHPLPGSystemState(inits)
        self.inputs = CHPLPGSystemInputs()

    def step(self):
        """Perform a simulation step"""

        err = None
        try:
            # First step the household
            self.heatdemand.inputs.step_size = self.inputs.step_size
            self.heatdemand.inputs.day_avg_t_air_deg_celsius = (
                self.inputs.day_avg_t_air_deg_celsius
            )
            self.heatdemand.step()
        except TypeError as te:
            err = te

        # Second step the chp
        p_set_kw = self._check_setpoint()
        self.chp.set_now_dt(self.inputs.now_dt)
        if self.inputs.e_th_demand_set_kwh is not None:
            self.chp.inputs.e_th_demand_set_kwh = (
                self.inputs.e_th_demand_set_kwh * self.config.e_th_demand_sign
            )
        elif err is None:
            self.chp.inputs.e_th_demand_set_kwh = (
                self.heatdemand.state.e_th_kwh
            )
        else:
            msg = (
                "One input is missing for CHP. Must at least provide "
                "one of 'day_avg_t_air_deg_celsius' or "
                "'e_th_demand_kwh'!"
            )
            LOG.error(msg)
            raise ValueError(msg)

        self.chp.set_step_size(self.inputs.step_size)
        self.chp.set_p_kw(p_set_kw * self.config.gsign)
        self.chp.step()

        # Update the references
        self.state.p_kw = self.chp.state.p_kw
        self.state.q_kvar = self.chp.state.q_kvar
        self.state.p_th_kw = self.chp.state.p_th_kw
        self.state.storage_t_c = self.chp.state.storage_t_c
        self.state.lubricant_l = self.chp.state.lubricant_l
        if self.inputs.e_th_demand_set_kwh is not None:
            self.state.e_th_demand_kwh = self.inputs.e_th_demand_set_kwh
        else:
            self.state.e_th_demand_kwh = self.heatdemand.state.e_th_kwh

        self.inputs.reset()

    def _check_setpoint(self):
        setpoint = self.inputs.p_set_kw
        if setpoint is not None:
            return abs(setpoint)
        hour = self.inputs.now_dt.hour
        default = self.config.default_schedule[hour]
        setpoint = self.config.p_max_kw * default / 100

        return abs(setpoint)

    def get_state(self):
        """Get state"""
        state_dict = {
            "household": self.heatdemand.get_state(),
            "chp": self.chp.get_state(),
        }
        return state_dict

    def set_state(self, state_dict):
        """Set state"""
        self.heatdemand.set_state(state_dict["household"])
        self.chp.set_state(state_dict["chp"])

        self.state.p_kw = self.chp.state.p_kw
        self.state.q_kvar = self.chp.state.q_kvar
        self.state.p_th_kw = self.chp.state.p_th_kw
        self.state.storage_t_c = self.chp.state.storage_t_c
        self.state.lubricant_l = self.chp.state.lubricant_l

    # @property
    # def set_percent(self):
    #     return self.chp.set_percent

    # @set_percent.setter
    # def set_percent(self, value):
    #     self.chp.set_percent = value
    #     self.inputs.p_set_kw = self.chp.inputs.p_set_kw



"""This module contains different DummyModels, which only implement the
the model interfaces. Their purpose is mainly for testing.

"""

from typing import Any, Dict

from pysimmods.model.config import ModelConfig
from pysimmods.model.inputs import ModelInputs
from pysimmods.model.model import Model
from pysimmods.model.state import ModelState


class DummyConfig(ModelConfig):
    def __init__(self, params: Dict[str, Any]):
        super().__init__(params)

        self.p_max_kw: float = params.get("p_max_kw", 500)
        self.p_min_kw: float = params.get("p_min_kw", 250)
        self.q_min_kvar: float = params.get("q_min_kvar", -550)
        self.q_max_kvar: float = params.get("q_max_kvar", 550)
        self.default_p_schedule = [375.0] * 24
        self.default_q_schedule = [125.0] * 24


class DummyState(ModelState):
    pass


class DummyInputs(ModelInputs):
    pass


class DummyModel(Model):
    """The dummy base model."""

    def __init__(self, params, inits):
        self.config = DummyConfig(params)
        self.state = DummyState(inits)
        self.inputs = DummyInputs()

    def step(self):
        self.state.p_kw = self.inputs.p_set_kw
        self.state.q_kvar = self.inputs.q_set_kvar
        if self.state.p_kw is None:
            self.state.p_kw = 0.0
        if self.state.q_kvar is None:
            self.state.q_kvar = 0.0

    def get_pn_max_kw(self):
        return self.config.p_max_kw

    def get_pn_min_kw(self):
        return self.config.p_min_kw

    def get_qn_max_kvar(self):
        return self.config.q_max_kvar

    def get_qn_min_kvar(self):
        return self.config.q_min_kvar

    def set_p_kw(self, p_kw: float) -> None:
        self.inputs.p_set_kw = p_kw

    def get_p_kw(self):
        return self.state.p_kw

    def set_q_kvar(self, q_kvar: float) -> None:
        self.inputs.q_set_kvar = q_kvar

    def get_q_kvar(self):
        return self.state.q_kvar

    # def set_percent(self, percentage):
    #     self.inputs.p_set_kw = (
    #         self.config.p_min_kw
    #         + percentage * (self.config.p_max_kw -
    # self.config.p_min_kw) / 100
    #     )

    # def get_percent_in(self) -> float:
    #     raise NotImplementedError()

    def get_percent_out(self) -> float:
        raise NotImplementedError()



"""
This module contains a household class that simulates thermal power demand
for the chp.

"""

import math
import random
from copy import deepcopy
from datetime import timedelta

from .config import HeatDemandConfig
from .inputs import HeatDemandInputs
from .state import HeatDemandState
from .util import hprofiles

DEMAND_PROFILES = dict()


class HeatDemand:
    """A simple model for thermal power demand.

    :constant CHP_AVERAGE_RUNTIME:
        source: 'IHK Projekte Hannover, Blockheizkraftwerke,
        Seite 5 IHK Projekte Durschnittliche Betriebsstunden
        eines BHKWs'
    :constant CHP_MIN_TH_CONS_COEF:
        source: 'IHK Projekte Hannover, Blockheizkraftwerke,
        Seite 5 Prozentzahl für die untere minimale thermische
        Leistung des Jahresverbrauchs eines Householdtypes'
    :constant CHP_MAX_TH_CONS_COEF:
        source: 'IHK Projekte Hannover, Blockheizkraftwerke,
        Seite 5 Prozentzahl für die obere maximale thermische
        Leistung des Jahresverbrauchs eines Householdtypes'
    :constant DEGREE_OF_EFFIENCY:
        source: 'Niedertemperatur- und Brennwertkessel.
        Wissenswertes über moderne Zentralheizungsanlagen.
        Hessisches Ministerium für Umwelt, Energie,
        Landwirtschaft und Verbraucherschutz. 2011'

    """

    CHP_AVERAGE_RUNTIME = 5_000.0
    CHP_MIN_TH_CONS_COEF = 1.0
    CHP_MAX_TH_CONS_COEF = 1.5
    # DEGREE_OF_EFFIENCY = 0.93

    def __init__(self, config, state):
        self.config = HeatDemandConfig(config)
        self.state = HeatDemandState(state)
        self.inputs = HeatDemandInputs()

    def step(self):
        """perform a simulation step"""

        next_state = deepcopy(self.state)

        # calculate allocation temperature for the current day
        # (weighted mean temperature of the current and the last
        # three days)
        t_alloc = (
            1.0 * self.inputs.day_avg_t_air_deg_celsius
            + 0.5 * self.state.t_last_1_deg_celsius
            + 0.25 * self.state.t_last_2_deg_celsius
            + 0.125 * self.state.t_last_3_deg_celsius
        ) / (1 + 0.5 + 0.25 + 0.125)

        # profile function 'SigLinDe': calculate daily consumption
        # depending on the allocation temperature. Has both a sigmoid
        # and a linear part
        tmp = math.pow(
            self.config.const_b / (t_alloc - self.config.const_v_0),
            self.config.const_c,
        )
        sigmoid = self.config.const_a / (1.0 + tmp) + self.config.const_d
        linear = max(
            self.config.const_m_h * t_alloc + self.config.const_b_h,
            self.config.const_m_w * t_alloc + self.config.const_b_w,
        )

        daily_cons = (
            self.config.consumer_const
            * (sigmoid + linear)
            * self.config.weekday_const[self.state.now_dt.weekday()]
        )

        next_state.e_th_kwh = (
            daily_cons
            * self.config.degree_of_efficiency
            * self.config.load_profile[self.state.now_dt.hour]
            * self.inputs.step_size
            / 3_600
        )

        next_state.now_dt += timedelta(seconds=self.inputs.step_size)
        if next_state.now_dt.day > self.state.now_dt.day:
            # we reached a new day
            next_state.t_last_3_deg_celsius = next_state.t_last_2_deg_celsius
            next_state.t_last_2_deg_celsius = next_state.t_last_1_deg_celsius
            next_state.t_last_1_deg_celsius = (
                self.inputs.day_avg_t_air_deg_celsius
            )

        self.state = next_state

    @property
    def e_th_kwh(self):
        """Returns the current thermal power demand"""
        return self.state.e_th_kwh

    def get_state(self) -> dict:
        """Return the current state of the model.

        Returns
        -------
        dict
            The current state of the model in form of a dictionary
            containing entries for all state variables. Returned dict
            can be assigned to the *inits* argument when creating a new
            model instance.

        """

        try:
            return {
                attr: getattr(self.state, attr)
                for attr in self.state.__slots__
            }
        except AttributeError:
            return deepcopy(self.state.__dict__)

    def set_state(self, state: dict) -> None:
        """Set the current state of the model.

        Parameters
        ----------
        state : dict
            A *dict* containing entries for all state variables.

        """
        for attr, value in state.items():
            setattr(self.state, attr, value)


def create_heatdemand(config, state):
    """Creates a household object"""

    if not DEMAND_PROFILES:
        load_demand_profiles()

    p_th_prod = -config.get("chp_p_th_prod_kw", -400)
    profile = find_demand_profile(p_th_prod)

    # adapt consumer constant of the household, so that power and
    # consumption better fit to each other
    profile["consumer_constant"] = p_th_prod * 24 * 2 / 3

    heatdemand_obj = HeatDemand(profile, state)

    return heatdemand_obj


def load_demand_profiles():
    """Write me"""
    global DEMAND_PROFILES
    DEMAND_PROFILES = dict()
    for name, func in hprofiles.__dict__.items():
        if "__" not in name:
            DEMAND_PROFILES[name] = func()


def find_demand_profile(chp_p_th_kw):
    """Write me"""
    min_consumption = dict()
    max_consumption = dict()

    for key, val in DEMAND_PROFILES.items():
        # average annual consumption per average operating
        # hours of the chp
        consumption = (
            val["consumer_constant"] * 365 / HeatDemand.CHP_AVERAGE_RUNTIME
        )
        # calculate lower power boundary
        min_consumption[key] = consumption * HeatDemand.CHP_MIN_TH_CONS_COEF
        # calculate upper power boundary
        max_consumption[key] = consumption * HeatDemand.CHP_MAX_TH_CONS_COEF

    possible_profiles = list()
    distances = dict()

    # Look for fitting households
    for key, val in DEMAND_PROFILES.items():
        if abs(chp_p_th_kw >= min_consumption[key]) and abs(
            chp_p_th_kw <= max_consumption[key]
        ):
            # Profile fits optimally
            possible_profiles.append(key)
            distances[key] = 0.0
        else:
            # Profile does not fit, calculate difference
            lower = abs(abs(chp_p_th_kw) - min_consumption[key])
            upper = abs(abs(chp_p_th_kw) - max_consumption[key])
            distances[key] = lower if lower < upper else upper

    if len(possible_profiles) == 1:
        # Exactly one matching profile
        demand_profile = DEMAND_PROFILES[possible_profiles[0]].copy()
    elif len(possible_profiles) > 1:
        # More than one matching profile, select randomly
        demand_profile = DEMAND_PROFILES[
            random.choice(possible_profiles)
        ].copy()
    else:
        # No matching profile, select the nearest one

        min_type = "one_family"
        min_val = float("Inf")

        for key, val in distances.items():
            if val < min_val:
                min_val = val
                min_type = key

        demand_profile = DEMAND_PROFILES[min_type].copy()
    return demand_profile



"""This module contains the base class for all pysimmods models."""

import copy
from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Optional, Union

from .config import ModelConfig
from .inputs import ModelInputs
from .state import ModelState


class Model(ABC):
    """Base class for pysimmods models.

    Normally, other models should not directly derive from this class;
    there are interface classes for consumers, buffers, and generators,
    as well.

    The *Model* class provides an interface for all models within this
    package. It takes two dictionaries at construction time, one for
    configuration parameters of the model and one to provide an initial
    state for the model. Those can be accessed via different attributes.

    Methods for accessing the most important attributes are provided as
    well as generic get/set state methods.

    All interface methods have in common that they consider the sign
    convention, i.e, in the (default) passive sign convention, power
    consumption is denoted with a positive sign while generation has a
    negative sign. Internally, all models use positive signs. The only
    exception mark the buffer models, which can both consume and
    generate power.

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters.
    inits : dict
        A *dict* containing the state variables.

    Attributes
    ----------
    config: :class:`.ModelConfig`
        A config object holding all the configuration parameters of
        this model. These do not change during simulation.
    state: :class:`.ModelState`
        A state object holding all variable parameters of this model.
        These values normally change during each step.
    inputs: :class:`.ModelInputs`
        An inputs object defining the inputs for this model. In each
        step, all the inputs need to be provided.

    """

    def __init__(self, params: dict, inits: dict) -> None:
        self.config: ModelConfig
        self.state: ModelState
        self.inputs: ModelInputs

    @abstractmethod
    def step(self) -> None:
        """Perform a simulation step.

        All required input variables need to be set beforehand. In
        *step*, inputs will be interpreted and calculation results of
        the model will be stored in the state of the model.
        """

    def get_state(self) -> dict:
        """Return the current state of the model.

        Returns
        -------
        dict
            The current state of the model in form of a dictionary
            containing entries for all state variables. Returned dict
            can be assigned to the *inits* argument when creating a new
            model instance.

        """

        try:
            return {
                attr: getattr(self.state, attr)
                for attr in self.state.__slots__
            }
        except AttributeError:
            return copy.deepcopy(self.state.__dict__)

    def set_state(self, state: dict) -> None:
        """Set the current state of the model.

        Parameters
        ----------
        state : dict
            A *dict* containing entries for all state variables.

        """
        for attr, value in state.items():
            setattr(self.state, attr, value)

    @abstractmethod
    def set_p_kw(self, p_kw: float) -> None:
        """Set the target value for active power for the model's next
        step.

        The value will be checked against the boundaries p_min_kw and
        p_max_kw.

        """

    @abstractmethod
    def get_p_kw(self) -> float:
        """Return the current active power output of the model."""

    def get_p_set_kw(self) -> Optional[float]:
        """Return the current setpoint for active power of the model."""
        return self.inputs.p_set_kw

    def get_pn_max_kw(self) -> float:
        """Maximum nominal active power output of the model.

        The output depends on the used sign convention. The base
        implementation is done in the consumer/generator/buffer
        subclasses.

        """
        return self.config.p_max_kw

    def get_pn_min_kw(self) -> float:
        """Minimum nominal active power output of the model.

        The output depends on the used sign convention. The minimum
        output is always the minimal output **while the model is
        active**. In contrast, the models can be turned off, which is
        indicated by an output of 0.

        """
        return self.config.p_min_kw

    def set_cos_phi(self, cos_phi: float) -> None:
        self.inputs.cos_phi = max(0, min(1, cos_phi))

    def get_cos_phi(self) -> Optional[float]:
        return self.state.cos_phi

    def set_percent(self, percentage: float) -> None:
        """Convenience function to set the desired output of the model.

        *percentage* is expected to be a value between 0 and 100 (unless
        the :attr:`_use_decimal_percentage` is set to True; in that case
        the value is expected to be within [0, 1.0]).

        A value of exactly 0 will indicate the model to turn off. For
        any other value, percentage is some valid value between the
        model's p_min_kw and p_max_kw (exception here, if the model is
        able to output reactive power, see (insert link)).

        Parameters
        ----------
        percentage: float
            Percentage setpoint value for the model between 0 and 100.
        """
        if percentage <= 0:
            self.inputs.p_set_kw = 0
            return

        if self.config.use_decimal_percent:
            decimal = max(min(abs(percentage), 1.0), 0.0)
        else:
            # Internally, decimal percentage is used.
            decimal = max(min(abs(percentage), 100.0), 0.0) / 100.0

        if decimal <= 0.0001:
            # Map values lower than 0.01 % to minimum power
            decimal = 0
        if decimal > 1.0:
            # Map values higher than 1 to maximum power
            decimal = 1.0

        p_max_kw = self.get_pn_max_kw()
        p_min_kw = self.get_pn_min_kw()

        self.inputs.p_set_kw = p_min_kw + decimal * (p_max_kw - p_min_kw)

    def get_percent_in(self) -> float:
        """Return the percentage setpoint."""
        p_kw = self.inputs.p_set_kw
        if p_kw is None:
            return None

        return self._get_percent(
            p_kw, self.get_pn_min_kw(), self.get_pn_max_kw()
        )

    def get_percent_out(self) -> float:
        """Return the percentage output power."""
        p_kw = self.state.p_set_kw
        if p_kw is None:
            return None

        return self._get_percent(
            p_kw, self.get_pn_min_kw(), self.get_pn_max_kw()
        )

    def set_step_size(self, step_size: int):
        step_size = int(step_size)
        assert step_size > 0, "step_size must be greater than zero"

        self.inputs.step_size = step_size

    def set_now_dt(self, now: Union[datetime, str, int]) -> None:
        """Set the current date and time of the model.

        The parameter *now* can be either a datetime object, an UTC ISO
        8601 time string or a unit timestamp (in seconds, as int). The
        value will be converted and stored as datetime in UTC time
        internally.

        Parameters
        ----------
        now: Union[datetime, str, int]
            The current date and time to set.

        """

        self.inputs.now_dt = now

    def get_now_dt(self) -> datetime:
        """Return the current date and time of the model.

        TODO: Return the value from inputs or from state?

        """
        raise NotImplementedError()

    def set_q_kvar(self, q_kvar: Optional[float]) -> None:
        """Set the target value for reactive power for the model's next
        step.

        The value will be checked against the boundaries q_min_kvar and
        q_max_kvar.

        """
        self.inputs.q_set_kvar = q_kvar

    def get_q_set_kvar(self) -> Optional[float]:
        return self.inputs.q_set_kvar

    def get_q_kvar(self) -> float:
        """Return the current reactive power output of the model."""
        return 0

    def get_qn_max_kvar(self) -> float:
        """Maximum nominal reactive power output of the model.

        The output depends on the used sign convention. The base
        implementation is done in the consumer/generator/buffer
        subclasses.

        """
        raise NotImplementedError()

    def get_qn_min_kvar(self) -> float:
        """Minimum nominal reactive power output of the model.

        The output depends on the used sign convention and the model's
        capability to output reactive power. The minimum output is
        always the minimal output **while the model is active**.

        In contrast, the models can be turned off, which is
        indicated by an output of 0 as well as if the model cannot
        output reactive power.

        """
        raise NotImplementedError()

    def get_default_p_schedule(self) -> List[float]:
        """Return the default schedule for active power.

        Returns
        -------
        list
            A *list* containing a default schedule for active power for
            each hour of the day (i.e., len(default_schedule) == 24).
            This is used if no other *p* input is provided.

        """
        return self.config.default_p_schedule

    def get_default_q_schedule(self) -> List[float]:
        """Return the default schedule for reactive power.

        Returns
        -------
        list
            A *list* containing a default schedule for reactive power
            for each hour of the day (i.e., len(default_q_schedule) ==
            24). This is used if no other *q* input is provided.

        """
        return self.config.default_p_schedule

    def _get_percent(self, val, val_min, val_max):
        if val > 0:
            # Following the primary purpose
            decimal = (val - val_min) / (val_max - val_min)
            if decimal < 0:
                # val is too low, turn off
                return 0

            # We don't allow 0 if the unit is turned on
            decimal = max(0.0001, decimal)

            if self.config.use_decimal_percent:
                return decimal
            else:
                return decimal * 100
        else:
            # Unit is off or does not follow primary purpose
            return 0

    def get_default_p_set(self, hour: int) -> Optional[float]:
        """Return the value of the default schedule at *hour*"""
        hour = max(0, min(23, hour))
        try:
            return self.config.default_p_schedule[hour]
        except TypeError:
            # There is no schedule
            return None

    def get_default_q_set(self, hour: int) -> Optional[float]:
        """Return the value of the default schedule at *hour*"""
        hour = max(0, min(23, hour))
        try:
            return self.config.default_q_schedule[hour]
        except TypeError:
            # There is no schedule
            return None



"""This module contains a :class:`mosaik_api.Simulator` for all models
of the pysimmods package.

"""

import json
import logging
import pprint
import warnings
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional

import mosaik_api_v3
import numpy as np
from midas.util.dict_util import strtobool
from midas.util.logging import set_and_init_logger
from midas.util.runtime_config import RuntimeConfig

from pysimmods.model.model import Model
from pysimmods.mosaik import LOG
from pysimmods.util.date_util import GER

from .meta import META, MODELS

VLOG = logging.getLogger("pysimmods.mosaik.verbose")


class PysimmodsSimulator(mosaik_api_v3.Simulator):
    """The Pysimmods simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid: str
        self.step_size: int
        self.now_dt: datetime
        self.key_value_logs: bool

        self.models: Dict[str, Model] = {}
        self.num_models: Dict[str, int] = {}

    def init(
        self,
        sid: str,
        start_date: str,
        step_size: int = 900,
        key_value_logs: Optional[bool] = None,
        time_resolution: float = 1.0,
        **kwargs,  # for compatibility
    ):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            The start date as UTC ISO 8601 date string.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.
        key_value_logs: bool, optional
            If set to True, a different, json-based logging format will
            be used. Default is False.
        time_resolution: float, optional
            Something new from mosaik 3.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).

        """
        self.sid = sid
        self.step_size = step_size
        self.now_dt = datetime.strptime(start_date, GER).astimezone(
            timezone.utc
        )

        self.key_value_logs = key_value_logs
        if self.key_value_logs is None:
            self.key_value_logs = RuntimeConfig().misc.get(
                "key_value_logs", False
            )

        return self.meta

    def create(
        self,
        num: int,
        model: str,
        params: Dict[str, Any],
        inits: Dict[str, Any],
    ):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num: int
            The number of models to create.
        model: str
            The name of the models to create. Must be present inside
            the simulator's meta.
        params: Dict[str, Any]
            The dictionary with parameters for the model to be created.
            Must be provided but can be empty.
        inits: Dict[str, Any]
            The dictionary with initial values for the state of the
            model to be created. Must be provided but can be empty.

        Returns
        -------
        list
            A list with information on the created entity.

        """
        entities = []
        self.num_models.setdefault(model, 0)

        for _ in range(num):
            eid = f"{model}-{self.num_models[model]}"
            self.models[eid] = MODELS[model](params, inits)
            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})

        return entities

    def step(
        self,
        time: int,
        inputs: Dict[str, Dict[str, Dict[str, Any]]],
        max_advance: int = 0,
    ):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.
        max_advance: int,
            Required for mosaik>=3 but not explicitly supported yet.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        if not self.key_value_logs:
            LOG.debug(
                "At step %d: Received inputs: %s.",
                time,
                pprint.pformat(inputs),
            )

        self._set_default_inputs()

        # Set inputs from other simulators
        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():
                if self.key_value_logs:
                    self._log_input(eid, attr, src_ids)

                # Use time information from time generator
                if attr == "local_time":
                    self._set_attr_local_time(eid, src_ids)
                    continue

                attr_sum = self._aggregate_attr(src_ids)
                self._set_remaining_attrs(eid, attr, attr_sum)

        # Step the models
        for model in self.models.values():
            model.step()

        # Update time for the next step
        self.now_dt += timedelta(seconds=self.step_size)

        return time + self.step_size

    def get_data(
        self, outputs: Dict[str, List[str]]
    ) -> Dict[str, Dict[str, Any]]:
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = {}
        for eid, attrs in outputs.items():
            for attr in attrs:
                value = self._get_remaining_attrs(eid, attr)
                data.setdefault(eid, dict())[attr] = value

                if self.key_value_logs:
                    self._log_output(eid, attr, value)

        if not self.key_value_logs:
            LOG.debug("Gathered outputs: %s.", pprint.pformat(data))

        return data

    def _set_default_inputs(self):
        VLOG.debug(
            "Setting step size %d and current time %s to all models.",
            self.step_size,
            self.now_dt,
        )
        for _, model in self.models.items():
            model.set_step_size(self.step_size)
            model.set_now_dt(self.now_dt)

    def _log_input(self, eid: str, attr: str, src_ids: Dict[str, Any]):
        for src_id, val in src_ids.items():
            if isinstance(val, np.ndarray):
                val = val.tolist()
            elif isinstance(val, np.integer):
                val = int(val)
            elif isinstance(val, (np.float32, np.floating)):
                val = float(val)

            LOG.debug(
                json.dumps(
                    {
                        "id": f"{self.sid}_{eid}",
                        "name": eid,
                        "model": eid.split("-")[0],
                        "type": "input",
                        "attribute": attr,
                        "source": src_id,
                        "value": val,
                    }
                )
            )

    def _log_output(self, eid: str, attr: str, value: Any):
        LOG.debug(
            json.dumps(
                {
                    "id": f"{self.sid}_{eid}",
                    "name": eid,
                    "model": eid.split("-")[0],
                    "type": "output",
                    "attribute": attr,
                    "value": value,
                }
            )
        )

    def _set_attr_local_time(self, eid: str, src_ids: Dict[str, Any]) -> bool:
        for val in src_ids.values():
            self.models[eid].set_now_dt(val)
            self.now_dt = datetime.strptime(val, GER).astimezone(timezone.utc)
            return True

        return False

    def _aggregate_attr(self, src_ids: Dict[str, Any]) -> float:
        """Aggregate inputs from different sources.

        If more inputs for one source exists, the average is calculated.

        """
        attr_sum = 0
        for val in src_ids.values():
            if val is None:
                continue
            if isinstance(val, (list, np.ndarray)):
                # This should only happen if palaestrAI is used
                val = val[0]
            attr_sum += val
        attr_sum /= len(src_ids)

        return float(attr_sum)

    # def _set_percent_power(self, eid: str, attr_sum: float):
    #     attr_sum /= self.percent_factor
    #     attr_sum = self.models[eid].get_pn_min_kw() + attr_sum * (
    #         self.models[eid].get_pn_max_kw() - self.models[eid].
    # get_pn_min_kw()
    #     )
    #     self.models[eid].set_p_kw(attr_sum)

    def _set_remaining_attrs(self, eid: str, attr: str, attr_sum: float):
        # Apply corrections
        if attr in ("p_set_mw", "p_th_set_mw", "q_set_mvar"):
            attr = attr.replace("m", "k")
            attr_sum *= 1e3

        # Set the inputs
        if attr == "set_percent":
            warnings.warn(
                "Using set_percent is deprecated and will be ignored. Use "
                "p_set_kw or p_set_mw instead",
                UserWarning,
            )

        elif attr == "p_set_kw":
            self.models[eid].set_p_kw(attr_sum)
        elif attr == "q_set_kvar":
            self.models[eid].set_q_kvar(attr_sum)
        else:
            setattr(self.models[eid].inputs, attr, attr_sum)

    def _get_remaining_attrs(self, eid: str, attr: str) -> float:
        # Apply correction of the attr if necessary
        if attr in ("p_mw", "p_th_mw", "q_mvar"):
            true_attr = attr.replace("m", "k")
        elif attr == "p_possible_max_mw":
            true_attr = "p_possible_max_kw"
        else:
            true_attr = attr

        if true_attr == "p_kw":
            value = self.models[eid].get_p_kw()
        elif true_attr == "q_kvar":
            value = self.models[eid].get_q_kvar()
        else:
            value = getattr(self.models[eid].state, true_attr)

        # Apply correction of the value if necessary
        if attr in ("p_mw", "p_th_mw", "q_mvar", "p_possible_max_mw"):
            value *= 1e-3

        return value


if __name__ == "__main__":
    set_and_init_logger(
        0, "pysimmods-logfile", "pysimmods-mosaik.log", replace=True
    )
    LOG.info("Starting mosaik simulation...")
    mosaik_api_v3.start_simulation(PysimmodsSimulator())



"""This module contains the state model of the Biogas plant."""

from pysimmods.model.state import ModelState


class BiogasState(ModelState):
    """State parameters of the biogas plant model.

    This class captures the state of the biogas model. States normally
    change during the step-method of this model.

    Parameters
    ----------
    inits: dict
        A dictionary containing initialization parameters for the
        biogas model. The :attr:`gas_fill_percent` is the only
        model-specific key that ist required.

    Attributes
    ----------
    gas_fill_percent: float
        Current level of the gas storage in [%].
    gas_prod_m3: float
        Amount of gas produced in current step in [m^3].
    gas_cons_m3: float
        Amount of gas consumed in the current step in [m^3].
    p_th_kw: float
        Thermal power output in [kW].
    gas_critical: bool, optional
        Is true, when storage boundaries are exceeded.
    burn_gas: bool, optional
        Is true, when the upper storage boundary is exceeded.
        Resets to False, when gas_fill_percent gets below 60 %.
    pool_gas: bool, optional
        Is true, when the lower storage boundary is exceeded.
        Resets to False, when gas_fill_percent gets above 40 %.

    """

    def __init__(self, inits):
        super().__init__(inits)
        self.gas_fill_percent = inits["gas_fill_percent"]
        self.gas_prod_m3 = inits.get("gas_prod_m3", None)
        self.gas_cons_m3 = inits.get("gas_cons_m3", None)
        self.p_th_kw = inits.get("p_th_kw", None)
        self.gas_critical = inits.get("gas_critical", False)
        self.burn_gas = inits.get("burn_gas", False)
        self.pool_gas = inits.get("pool_gas", False)



import math

import numpy as np
import pandas as pd

M_AIR_KG_PER_MOL: float = 0.029
ACCELERATION_GRAVITY_M_PER_S2: float = 9.81
GAS_CONSTANT_J_PER_MOL_K: float = 8.314
GAS_CONSTANT_J_PER_KG_K: float = 287.058
T_0_DEGREE_KELVIN: float = 288.15
P_0_HPA: float = 1013.25
P_GRADIENT_HPA_PER_M: float = 1 / 8
AIR_DENSITY_KG_PER_M3: float = 1.225


# # Calculate the pressure at altitude h using the barometric formula
# def atmospheric_pressure(
#     h, P0=1013.25, h0=0
# ):  # P0=1013.25   Standard sea level pressure (Pa)
#     # M = 0.029  # Molar mass of air in kg/mol
#     # g = 9.81  # Acceleration due to gravity in m/s^2
#     # R = 8.314  # Universal gas constant in J/(mol·K)
#     # T0 = 288.15  # Standard temperature at sea level in K

#     # Calculate the temperature at altitude h using the standard lapse rate
#     lapse_rate = -0.0065  # K/m
#     T = T0 + lapse_rate * h

#     # Calculate atmospheric pressure using the barometric formula
#     P = P0 * math.exp(-M * g * (h - h0) / (R * T))

#     return P


def barometric(
    pressure_hpa: float,
    temperature_at_height: float,
    data_height: float,
    target_height: float,
):
    """
    the barometric height equation to calculate the air density at hub
    height using pressure and Temperature at hub height.
    R is the specific gas constant for dry air (around
    287.05 J/(kg·K)).
    Pressure gradient of -1/8 hPa/m   ASSUMPTION
        1.225   # rho
        pressure:  Air pressure in Pa.
        pressure_height : Height in m for which the parameter `pressure`
            applies. the  same for  temperature_height
        pressure_height: hub height of wind turbine in m.
        temperature_hub_height : Air temperature at hub height in K
    """

    #

    # pressure_height = temperature_height
    # pressure = self.atmospheric_pressure(
    #      pressure
    # )
    # pressure_height = self.atmospheric_pressure(hub_height )
    #   temperature_hub_height = self.linear_gradient(
    #      temperature, self.config.temperature_height, self.config.hub_height
    #  )

    return (
        pressure_correction(pressure_hpa, data_height, target_height)
        * AIR_DENSITY_KG_PER_M3
        * T_0_DEGREE_KELVIN
        / (P_0_HPA * temperature_at_height)
    )


def ideal_gas(
    pressure_hpa: float,
    temperature_at_height: float,
    data_height: float,
    target_height: float,
):
    """
    The barometric height equation to calculate the air density at hub
    height using pressure and Temperature at hub height
    gas constant of dry air (287.058 J/(kg*K))
    return air density at hub height
    """
    #
    # pressure_height = self.config.temperature_height
    # temperature_hub_height = self.linear_gradient(temperature)

    return (
        pressure_correction(pressure_hpa, data_height, target_height)
        / (GAS_CONSTANT_J_PER_KG_K * temperature_at_height)
        * 100
    )


def pressure_correction(
    pressure_hpa: float, data_height: float, target_height: float
):
    return pressure_hpa - (target_height - data_height) * P_GRADIENT_HPA_PER_M



"""This module contains the config information for the heat demand
model.

"""


class HeatDemandConfig:
    """Captures the configuration variables of the heat demand model."""

    def __init__(self, params):
        self.const_a = params.get("A")
        self.const_b = params.get("B")
        self.const_c = params.get("C")
        self.const_d = params.get("D")
        self.const_v_0 = params.get("V_0")
        self.const_m_h = params.get("M_H")
        self.const_b_h = params.get("B_H")
        self.const_m_w = params.get("M_W")
        self.const_b_w = params.get("B_W")

        self.load_profile = params.get("load_profile")
        self.consumer_const = params.get("consumer_constant")
        self.weekday_const = params.get("weekday_constants")

        self.degree_of_efficiency = 0.93



"""This module contains the state model of the chp."""

from pysimmods.model.state import ModelState


class CHPCNGState(ModelState):
    """A CHP CNG state information class."""

    def __init__(self, init_vals):
        super().__init__(init_vals)

        self.active_s = init_vals["active_s"]
        """Time since unit switched to current power level in [s]."""
        self.active_s_per_day = init_vals["active_s_per_day"]
        """Time the unit is active on the current day in [s]."""
        self.inactive_s = init_vals["inactive_s"]
        """Time since unit is switched off in [s]."""
        self.inactive_s_per_day = init_vals["inactive_s_per_day"]
        """Time the unit is inactive on the current day in [s]."""

        self.restarts = init_vals["restarts"]
        """Number of restarts on the current day."""

        self.p_th_kw = None
        """Thermal power generated in [kW]."""
        self.gas_cons_m3 = None
        """Gas consumed in the last step in [m^3]."""



"""This module contains the *mosaik_api* meta definition for pysimmods
models.
"""

from pysimmods.buffer.batterysim import Battery
from pysimmods.consumer.hvacsim import HVAC
from pysimmods.generator import WindPowerPlantSystem
from pysimmods.generator.biogassim import BiogasPlant
from pysimmods.generator.chplpgsystemsim import CHPLPGSystem
from pysimmods.generator.dieselsim import DieselGenerator
from pysimmods.generator.pvsystemsim import PVPlantSystem

_SHARED_ATTRS = [
    "p_set_mw",
    "p_set_kw",
    "p_mw",
    "p_kw",
    "q_mvar",
    "q_kvar",
    "local_time",
]


META = {
    "type": "time-based",
    "models": {
        "Battery": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["soc_percent"] + _SHARED_ATTRS,
        },
        "Photovoltaic": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "q_set_mvar",
                "q_set_kvar",
                "t_air_deg_celsius",
                "t_module_deg_celsius",
                "bh_w_per_m2",
                "dh_w_per_m2",
                "s_module_w_per_m2",
                "inverter_inductive",
                "p_possible_max_mw",
            ]
            + _SHARED_ATTRS,
        },
        "CHP": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "day_avg_t_air_deg_celsius",
                "p_th_mw",
                "e_th_demand_set_kwh",
                "e_th_demand_kwh",
            ]
            + _SHARED_ATTRS,
        },
        "HVAC": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["t_air_deg_celsius"] + _SHARED_ATTRS,
        },
        "Biogas": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["p_th_mw"] + _SHARED_ATTRS,
        },
        "DieselGenerator": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": _SHARED_ATTRS,
        },
        "WindTurbine": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "wind_v_m_per_s",
                "t_air_deg_celsius",
                "air_pressure_hpa",
            ]
            + _SHARED_ATTRS,
        },
    },
}

MODELS = {
    "Photovoltaic": PVPlantSystem,
    "HVAC": HVAC,
    "Battery": Battery,
    "CHP": CHPLPGSystem,
    "DieselGenerator": DieselGenerator,
    "Biogas": BiogasPlant,
    "WindTurbine": WindPowerPlantSystem,
}



from pysimmods.model.config import ModelConfig
from pysimmods.model.generator import Generator
from pysimmods.model.inputs import ModelInputs
from pysimmods.model.state import ModelState


class DummyGeneratorConfig(ModelConfig):
    def __init__(self, params):
        params.setdefault("default_schedule", [10] * 24)
        super().__init__(params)

        self.p_max_kw = params.get("p_max_kw", 500)
        self.p_min_kw = params.get("p_min_kw", 250)


class DummyGeneratorState(ModelState):
    pass


class DummyGeneratorInputs(ModelInputs):
    pass


class DummyGenerator(Generator):
    """The dummy buffer model."""

    def __init__(self, params, inits):
        self.config = DummyGeneratorConfig(params)
        self.state = DummyGeneratorState(inits)
        self.inputs = DummyGeneratorInputs()

    def step(self):
        self.state.p_kw = abs(self.inputs.p_set_kw)
        if self.state.p_kw != 0:
            self.state.p_kw = min(
                self.config.p_max_kw,
                max(self.config.p_min_kw, self.state.p_kw),
            )



"""This module contains the input model for the Wind Turbine System."""

from typing import Optional

from ...model.inputs import ModelInputs


class WindSystemInputs(ModelInputs):
    def __init__(self):
        super().__init__()

        self.cos_phi_set: Optional[float] = None
        self.inverter_inductive: Optional[bool] = None

        self.t_air_deg_celsius: Optional[float] = None
        self.wind_v_m_per_s: Optional[float] = None
        self.air_pressure_hpa: Optional[float] = None

        # Not used yet
        self.wind_dir_deg: Optional[float] = None



"""This module contains a simulation model of a PV plant. The model was
developed within the scope of the research project iHEM (intelligent
Home Energy Management).

"""

from pysimmods.generator.pvsim.config import PVConfig
from pysimmods.generator.pvsim.inputs import PVInputs
from pysimmods.generator.pvsim.state import PVState
from pysimmods.model.generator import Generator
from pysimmods.util import irradiance, solargeometry


class PhotovoltaicPowerPlant(Generator):
    """Simulation model of a photovoltaic plant.

    A PV plant consists of one or more modules, but the model treats
    all modules as one big module. Impact of the temperature of the PV
    module on efficiency is considered. Therefore the temperature of
    the PV plant is explicitly modelled. The PV modules are considered
    to be homogeneous bodies, which are specified by their mass per
    unit area and heat capacity. To model the temperature of the PV
    modules heat flows between the PV modules and environment are
    calculated in each simulation step. The PV modules are heated by
    the share of sun radiation, which is not reflected by the surface
    of the PV modules and which is not converted into electric energy.
    At the same time, the PV plant releases heat to the ambient air.
    This heat flow is considered to be proportional to the temperature
    gradient between air temperature and module temperature. The
    corresponding proportionality factor is the heat transmission
    coefficient between air and module. Wind speed dependency of this
    coefficient is not taken into account.

    You have to provide the two dictionaries *params* and *inits*.
    The dict *params* provides the parameters for the PV plant model
    and should at least look like this::

        {"a_m2": 15.0, "eta_percent": 25}

    The parameter *a_m2* specifies the overall area of the PV modules
    in square meter. Additional parameters can be provided, see
    :class:`~.config.PVConfig`.
    The parameter *eta_percent* is optional and specifies the
    efficiency of the PV plant.
    The peak power of the PV plant depends directly on the size of the
    Surface and the efficiency and is defined as::

        p_peak_kw = eta_percent / 100 * a_m2

    The dict *inits* provides the initial values for the state variables.
    The PV plant model has one state variable that must be specified when
    the model is initialized. It is *t_module_deg_celsius* (indicates the
    temperature of the PV modules in °C)::

            {"t_module_deg_celsius": 25}

    Parameters
    ----------
    params : dict
        Configuration parameters as described above. See
        :class:`~.PVConfig` for all parameters.
    inits : dict
        Initialization parameters as described above. See
        :class:`~.PVState` for all parameters.

    Attributes
    ----------
    config : :class:`~.PVConfig`
        Stores the configuration parameters of the PV plant model.
    state : :class:`~.PVState`
        Stores the initialization parameters of the PV plant model.
    inputs : :class:`~.PVInputs`
        Stores the input parameters for each step of the PV plant
        model.

    """

    def __init__(self, params, inits):
        self.config = PVConfig(params)
        self.state = PVState(inits)
        self.inputs = PVInputs()

    def step(self):
        """perform simulation step

        Irradiance on plant depends on current solar position and
        orientation and tilt of the PV modules. To calculate the total
        irradiance on the tilted PV modules, functions
        :func:`~.tilted_surface` and :func:`~.compute_sun_position` are
        applied.

        """
        self._check_inputs()

        if self.config.has_external_irradiance_model:
            irradiance_w_per_m2 = self.inputs.s_module_w_per_m2
        else:
            if self.inputs.bh_w_per_m2 > 2 or self.inputs.dh_w_per_m2 > 2:
                sun_pos = solargeometry.compute_sun_position(
                    self.inputs.now_dt.timestamp(),
                    self.config.lat_deg,
                    self.config.lon_deg,
                )
                args = [
                    self.inputs.bh_w_per_m2,
                    self.inputs.dh_w_per_m2,
                    self.config.tilt_deg,
                    self.config.orient_deg,
                    sun_pos["elevation"],
                    sun_pos["azimuth"],
                ]
                irradiance_w_per_m2 = irradiance.tilted_surface(*args)
            else:
                irradiance_w_per_m2 = 0

        # calculate heat balance, irradiance induced heat flow
        # (Watt/m2, pos. sign indicates heat uptake of plant)
        if not self.config.is_static_t_module:
            t_module_change = self._calculate_module_temperature(
                irradiance_w_per_m2
            )

            self.state.t_module_deg_celsius += t_module_change

        # temperature for which eta_percent was measured
        t_ref_deg_celsius = 25  # °C

        # calculate alternating current power output of plant
        k_temp = (
            1
            - self.config.beta_percent_per_kelvin
            * max(self.state.t_module_deg_celsius - t_ref_deg_celsius, 0)
            / 100
        )
        # factor for temperature dependency of module efficiency

        eff_rel_max = (self.config.k_m_w_per_m2 + 1e3) / 1e3
        k_irradicance = (
            eff_rel_max
            * irradiance_w_per_m2
            / (self.config.k_m_w_per_m2 + irradiance_w_per_m2)
        )
        # factor for irradiance dependency of module efficiency
        # k_irr is modeled by a saturation curve whose function value is 0.5 if
        # the total irradiance equals K_m and 1.0 if the total
        # irradiance equals the irradiance at standard conditions
        # (1000 Watt/m2)

        self.state.p_kw = (
            irradiance_w_per_m2
            * self.config.eta_percent
            / 100
            * k_temp
            * k_irradicance
            * self.config.a_m2
            / 1e3
        )
        # print(
        #     f"Now: {self.inputs.now_dt.timestamp()} -- "
        #     f"{irradiance_w_per_m2} * {self.config.eta_percent}"
        #     f"/ 100 * {k_temp} * {k_irradicance} * {self.config.a_m2} / 1e3"
        #     f" = {self.state.p_kw} kW"
        # )

    def _check_inputs(self):
        error = self.inputs.step_size is None
        error = error or self.inputs.t_air_deg_celsius is None
        if not self.config.has_external_irradiance_model:
            error = error or self.inputs.bh_w_per_m2 is None
            error = error or self.inputs.dh_w_per_m2 is None
            error = error or self.inputs.now_dt is None

        if error:
            raise KeyError("At least one input is not set")

    def _calculate_module_temperature(self, irradiance_w_per_m2):
        if self.inputs.step_size > self.config.max_step_size:
            steps = self.inputs.step_size // self.config.max_step_size
            step_sizes = [self.config.max_step_size] * steps
            remainder = self.inputs.step_size % self.config.max_step_size
            if remainder > 0:
                step_sizes.append(remainder)

        else:
            step_sizes = [self.inputs.step_size]

        t_module = self.state.t_module_deg_celsius

        for step_size in step_sizes:
            q_dot_irradiance = irradiance_w_per_m2 * (
                1
                - self.config.eta_percent / 100
                - self.config.reflex_percent / 100
            )
            # heat flow induced by temperature gradient between module and air
            q_dot_temp_grad = self.config.alpha_w_per_m2_kelvin * (
                self.inputs.t_air_deg_celsius - t_module
            )
            # heat transfer in current time step [J/m2]
            e_j_per_m2 = (q_dot_irradiance + q_dot_temp_grad) * step_size
            # calculate new temperature of module
            t_module += e_j_per_m2 / (
                self.config.rho_kg_per_m2 * self.config.c_j_per_kg_kelvin
            )
        return t_module - self.state.t_module_deg_celsius



"""This module contains the input model for the PV system."""

from pysimmods.model.inputs import ModelInputs


class PVSystemInputs(ModelInputs):
    """Input variables for the PV plant system.

    Attributes
    ----------
    cos_phi_set : float
        See :attr:`~.InverterInputs.cos_phi_set`.
    inverter_inductive:: bool
        See _attr:`~.InverterInputs.inverter_inductive`.
    bh_w_per_m2 : float
        See :attr:`~.PVInputs.bh_w_per_m2`.
    dh_w_per_m2 : float
        See :attr:`~.PVInputs.dh_w_per_m2`.
    s_module_w_per_m2 : float
        See :attr:`~.PVInputs.s_module_w_per_m2`.
    t_air_deg_celsius : float
        See :attr:`~.PVInputs.t_air_deg_celsius`.


    """

    def __init__(self):
        super().__init__()

        self.cos_phi_set = None
        self.inverter_inductive = None
        self.bh_w_per_m2 = None
        self.dh_w_per_m2 = None
        self.s_module_w_per_m2 = None
        self.t_air_deg_celsius = None



"""Configuration of "TK-Anlagen" based on
https://www.tis-gdv.de/tis/tagungen/svt/svt10/weilhart/inhalt-htm/

"""

import sys


def hvac_preset(pn_max_kw, **kwargs):
    params = _get_dict(pn_max_kw, "params")
    inits = _get_dict(pn_max_kw, "init")

    return params, inits


def _get_dict(p_kw, dicttype):
    thismodule = sys.modules[__name__]

    if p_kw in (0.08,):
        method = "hvac_{}w_{}".format(int(p_kw * 1000), dicttype)
    elif p_kw in (2, 60, 100, 230, 343, 1279):
        method = "hvac_{}kw_{}".format(int(p_kw), dicttype)
    else:
        # TODO: find
        raise ValueError(f"Not supported model size {p_kw} p_kw")

    return getattr(thismodule, method)()


def hvac_80w_params():
    """The classic fridge from practical training
    energy informatics.
    """
    params = {
        "p_max_kw": 0.08,
        "eta_percent": 200.0,
        "length": 0.5,
        "width": 0.5,
        "height": 0.5,
        "d_m": 0.02,
        "lambda_w_per_m_k": 0.019,
        "t_min_deg_celsius": 3.0,
        "t_max_deg_celsius": 7.0,
    }
    return params


def hvac_80w_init():
    """The classic fridge from practical training
    energy informatics.
    """
    inits = {
        "mass_kg": 5.0,  # eggs
        "c_j_per_kg_k": 3.18 * 1e3,  # eggs not frozen
        "theta_t_deg_celsius": 4.0,
        "cooling": True,
        "mode": "auto",
    }
    return inits


def hvac_2kw_params():
    """The classic air conditioning from practical
    training energy informatics."""
    params = {
        "p_max_kw": 2.0,
        "eta_percent": 200.0,
        "length": 4.0,
        "width": 5.0,
        "height": 2.5,
        "d_m": 0.25,
        "lambda_w_per_m_k": 0.5,
        "t_min_deg_celsius": 17.0,
        "t_max_deg_celsius": 23.0,
    }
    return params


def hvac_2kw_init():
    """The classic air conditioning from practical
    training energy informatics."""
    inits = {
        "mass_kg": 500.0,
        "c_j_per_kg_k": 2390.0,
        "theta_t_deg_celsius": 21.0,
        "cooling": True,
        "mode": "auto",
    }
    return inits


def hvac_60kw_params():
    """Following the information from the link in the module
    header, this cooling room has an energetic value of approx.
    210 kWh/(m^3*Year) and a storage volume of 800 m^3.

    (All calculations use rounded values)

    This results in a total of 168000 kWh/Year and 460 kWh/Day.

    Assuming a workload of 33%, this cooling house could consume
    up to 1380 kWh/Day maximum, and therefore 57.5 kWh/h.

    The thaw and cool factor allow a bit control over the cooling and
    thawing process. They are set to match the characteristics above.
    """
    params = {
        "p_max_kw": 60,
        "eta_percent": 10.25,
        "length": 20.0,
        "width": 8.0,
        "height": 5.0,
        "d_m": 0.2,
        "lambda_w_per_m_k": 0.12,
        "t_min_deg_celsius": -30.0,
        "t_max_deg_celsius": -18.0,
        "thaw_factor": 5,
        "cool_factor": 1.0,
    }
    return params


def hvac_60kw_init():
    """We calculate the the mass of content in relation to the
    classic air conditioning, which assumes 500kg of oak with a
    specific heat capacity of 2390 J/(kg*K). This goes into the
    equation as 500 kg * 2390 J/(kg*K) = 1195000 J/K.

    However, the cooling house does not store oak, but food.
    Non-froozen food, e.g. apples, have a specific heat capacity
    of 3890 J/(kg*K). To get a similar relation, we calculate
    1195000 J/K / 3890 J/(kg*K) = 311kg, and, subsequently,
    apply the relation of the cooling volumen, which is 50m^3
    for the air conditioning and 800m^3 for the cooling house,
    i.e.: 311 kg * 800 m^3 / 50 m^3 = 4966 kg of food content.
    """
    inits = {
        "mass_kg": 5000.0,
        "c_j_per_kg_k": 3850.0,
        "theta_t_deg_celsius": -22,
        "cooling": True,
        "mode": "auto",
    }
    return inits


def hvac_100kw_params():
    """See explanation of hvac_60kw_params.

    Energetic value: 160 kWh/(m^3*Year)
    Volume: 1800 m^3
    Workload: 33 %
    """
    params = {
        "p_max_kw": 100,
        "eta_percent": 45.0,
        "length": 30.0,
        "width": 10.0,
        "height": 6.0,
        "d_m": 0.2,
        "lambda_w_per_m_k": 0.12,
        "t_min_deg_celsius": -30.0,
        "t_max_deg_celsius": -18.0,
        "thaw_factor": 10,
        "cool_factor": 0.1375,
    }
    return params


def hvac_100kw_init():
    """See explanation of hvac_60kw_init."""
    inits = {
        "mass_kg": 11_200.0,
        "c_j_per_kg_k": 3850.0,
        "theta_t_deg_celsius": -29,
        "cooling": True,
        "mode": "auto",
    }
    return inits


def hvac_230kw_params():
    """See explanation of hvac_60kw_params.

    Energetic value: 140 kWh/(m^3*Year)
    Volume: 4800 m^3
    Workload: 33 %
    """
    params = {
        "p_max_kw": 230,
        "eta_percent": 45.0,
        "length": 40.0,
        "width": 15.0,
        "height": 8.0,
        "d_m": 0.2,
        "lambda_w_per_m_k": 0.14,
        "t_min_deg_celsius": -30.0,
        "t_max_deg_celsius": -18.0,
        "thaw_factor": 13,
        "cool_factor": 0.07,
    }
    return params


def hvac_230kw_init():
    """See explanation of hvac_60kw_init."""
    inits = {
        "mass_kg": 29_856.0,
        "c_j_per_kg_k": 3850.0,
        "theta_t_deg_celsius": -19,
        "cooling": True,
        "mode": "auto",
    }
    return inits


def hvac_343kw_params():
    """See explanation of hvac_60kw_params.

    Energetic value: 120 kWh/(m^3*Year)
    Volume: 10000 m^3
    Workload: 40 %
    """
    params = {
        "p_max_kw": 343,
        "eta_percent": 45.0,
        "length": 50.0,
        "width": 25.0,
        "height": 8.0,
        "d_m": 0.2,
        "lambda_w_per_m_k": 0.15,
        "t_min_deg_celsius": -30.0,
        "t_max_deg_celsius": -18.0,
        "thaw_factor": 15.0,
        "cool_factor": 0.041,
    }
    return params


def hvac_343kw_init():
    """See explanation of hvac_60kw_init."""
    inits = {
        "mass_kg": 62_200.0,
        "c_j_per_kg_k": 3850.0,
        "theta_t_deg_celsius": -26,
        "cooling": True,
        "mode": "auto",
    }
    return inits


def hvac_1279kw_params():
    """See explanation of hvac_60kw_params.

    Energetic value: 70 kWh/(m^3*Year)
    Volume: 72000 m^3
    Workload: 45 %

    * 5040000 kWh/Year
    * 13808 kWh/Day (45%)
    * 30684 kWh/Day (100%)
    * 1279 kWh/h (100%)

    The efficiency is experimentally determined to match the values
    above (see the test file not_a_test_hvac).

    """
    params = {
        "p_max_kw": 1279,
        "eta_percent": 45.0,
        "length": 100.0,
        "width": 60.0,
        "height": 12.0,
        "d_m": 0.2,
        "lambda_w_per_m_k": 0.15,
        "t_min_deg_celsius": -30.0,
        "t_max_deg_celsius": -18.0,
        "thaw_factor": 50.0,  # 0.00235,
        "cool_factor": 0.011,
    }
    return params


def hvac_1279kw_init():
    """See explanation of hvac_60kw_init."""
    inits = {
        "mass_kg": 448_000.0,
        "c_j_per_kg_k": 3850.0,
        "theta_t_deg_celsius": -20,
        "cooling": True,
        "mode": "auto",
    }
    return inits



"""Presets for the battery model."""


def battery_preset(pn_max_kw, **kwargs):
    params = _get_dict(pn_max_kw, cap_kwh=kwargs.get("cap_kwh", pn_max_kw * 5))
    inits = {"soc_percent": kwargs.get("soc_percent", 50.0)}

    return params, inits


def _get_dict(pn_max_kw, cap_kwh):
    params = {
        "cap_kwh": cap_kwh,
        "soc_min_percent": 15,
        "p_charge_max_kw": pn_max_kw,
        "p_discharge_max_kw": pn_max_kw,
        "eta_pc": [-2.109566, 0.403556, 97.110770],
    }
    return params
